"""
driving_cycles.py loads a driving cycle based on
the name specific by the user.
The driving cycle returned is a numpy
array with speed levels (in km/h) for each
second of driving.
"""

import sys
from pathlib import Path
from typing import List, Tuple

import numpy as np
import yaml

from . import DATA_DIR

FILEPATH_DC_SPECS = DATA_DIR / "driving cycle" / "dc_specs.yaml"


def detect_vehicle_type(vehicle_sizes: List[str]) -> str:
    """
    Detect the type of vehicle based on the size of the vehicle.
    """

    dc = get_driving_cycle_specs()

    for vehicle_type in dc["columns"]:
        for dc_name in dc["columns"][vehicle_type]:
            for size in dc["columns"][vehicle_type][dc_name]:
                if size in vehicle_sizes:
                    return vehicle_type

    raise ValueError("The vehicle size is not in the list of available vehicle sizes.")


def get_driving_cycle_specs() -> dict:
    """Get driving cycle specifications.

    :returns: A dictionary with driving cycle specifications.
    :rtype: dict

    """

    with open(FILEPATH_DC_SPECS, "r") as f:
        return yaml.safe_load(f)


def get_dc_column_number(
    vehicle_type: str, vehicle_size: List[str], dc_name: str
) -> List[int]:
    """
    Loads YAML file that contains the column number.
    Return the column number given a vehicle type and driving cycle name.
    """

    dc_specs = get_driving_cycle_specs()

    if isinstance(vehicle_size, str):
        vehicle_size = [vehicle_size]

    if vehicle_type not in dc_specs["columns"]:
        raise KeyError(
            f"Vehicle type {vehicle_type} is not in the list of "
            f"available vehicle types: {list(dc_specs['columns'].keys())}"
        )

    if dc_name not in dc_specs["columns"][vehicle_type]:
        raise KeyError(
            f"Driving cycle {dc_name} is not in the list of "
            f"available driving cycles: {list(dc_specs['columns'][vehicle_type].keys())}"
        )

    if not all(
        vehicle in dc_specs["columns"][vehicle_type][dc_name]
        for vehicle in vehicle_size
    ):
        raise KeyError(
            f"Vehicle size(s) {vehicle_size} is not in the list of "
            f"available vehicle sizes: {list(dc_specs['columns'][vehicle_type][dc_name].keys())}"
        )

    return [dc_specs["columns"][vehicle_type][dc_name][s] for s in vehicle_size]


def get_data(
    filepath: Path, vehicle_type: str, vehicle_sizes: List[str], name: str
) -> np.ndarray:
    try:
        col = get_dc_column_number(vehicle_type, vehicle_sizes, name)
        arr = np.genfromtxt(filepath, delimiter=";")
        # we skip the headers
        dc = arr[1:, col]
        return dc

    except KeyError as err:
        print(err, "The specified driving cycle could not be found.")
        raise


def get_standard_driving_cycle_and_gradient(
    vehicle_type: str, vehicle_sizes: List[str], name: str
) -> Tuple[np.ndarray, np.ndarray]:
    """Get driving cycle and gradient data as a Pandas `Series`.

    Driving cycles are given as km/h per second up to 3200 seconds.

    :param name: The name of the driving cycle.
    e.g., WLTC (Worldwide harmonized Light vehicles Test Cycles)
    :type name: str

    :returns: A pandas DataFrame object with driving time
    (in seconds) as index, and velocity (in km/h) as values.
    :rtype: panda.Series

    """

    filepath_dc = DATA_DIR / "driving cycle" / f"{vehicle_type}.csv"
    filepath_gradient = DATA_DIR / "gradient" / f"{vehicle_type}.csv"

    # definition of columns to select in the CSV file
    # each column corresponds to a size class
    # since the driving cycle is simulated for each size class
    return (
        get_data(filepath_dc, vehicle_type, vehicle_sizes, name),
        get_data(filepath_gradient, vehicle_type, vehicle_sizes, name),
    )



"""
background_systems.py contains BackgroundSystem, which contains different arrays relating to
the evolution of fuel blends, sulfur content in fuels, and electricity mixes in different
countries over time.
"""


from typing import Dict, List

import numpy as np
import pandas as pd
import xarray as xr
import yaml

from . import DATA_DIR


def data_to_dict(csv_list: list) -> dict:
    """
    Returns a dictionary from a sequence of items.
    :param csv_list: list
    :return: dict
    """

    (_, *header), *data = csv_list
    csv_dict = {}
    for row in data:
        key, *values = row
        csv_dict[key] = dict(zip(header, values))

    return csv_dict


def get_electricity_losses() -> Dict[str, float]:
    """
    Retrieve cumulative electricity losses from high to medium and low voltage.
    Source: `ecoinvent v.3.6 <https://www.ecoinvent.org/>`_.

    :returns: dictionary
    :rtype: dict

    """
    filename = "cumulative_electricity_losses.csv"
    filepath = DATA_DIR / "electricity" / filename
    if not filepath.is_file():
        raise FileNotFoundError(
            "The CSV file that contains electricity " "mixes could not be found."
        )
    with open(filepath, encoding="utf-8") as file:
        csv_list = [[val.strip() for val in r.split(";")] for r in file.readlines()]

    return data_to_dict(csv_list)


def get_electricity_mix() -> xr.DataArray:
    """
    Retrieve electricity mixes and shape them into a xarray.
    Source:
        * for European countries (`ENTSOE TYNDP 2020 scenarios <https://2020.entsos-tyndp-scenarios.eu/>`_),
        * for African countries (`TEMBA <http://www.osemosys.org/temba.html>`_ model)
        * and for other countries (`IEA World Energy outlook 2017 <https://www.iea.org/reports/world-energy-outlook-2017>`_)

    :returns: An axarray with 'country' and 'year' as dimensions
    :rtype: xarray.core.dataarray.DataArray

    """
    filename = "electricity_mixes.csv"
    filepath = DATA_DIR / "electricity" / filename
    if not filepath.is_file():
        raise FileNotFoundError(
            "The CSV file that contains " "electricity mixes could not " "be found."
        )

    dataframe = pd.read_csv(filepath, sep=";", index_col=["country", "year"])
    dataframe = dataframe.reset_index()

    array = (
        dataframe.melt(id_vars=["country", "year"], value_name="value")
        .groupby(["country", "year", "variable"])["value"]
        .mean()
        .to_xarray()
    )
    array = array.interpolate_na(
        dim="year", method="linear", fill_value="extrapolate"
    ).clip(0, 1)
    array /= array.sum(axis=2)

    return array


def get_biofuel_share(filepath) -> xr.DataArray:
    """
    :return: Returns a xarray with share of biodiesel
    in the fuel blend, per country, per year.
    """
    if not filepath.is_file():
        raise FileNotFoundError(
            "The CSV file that contains biofuel shares " "shares could not be found."
        )
    dataframe = pd.read_csv(filepath, sep=";")

    return dataframe.groupby(["country", "year"]).sum().to_xarray().to_array()


def get_sulfur_content_in_fuel() -> xr.DataArray:
    """
    Retrieve sulfur content per kg of petrol and diesel.
    For CH, DE, FR, AU and SE, the concentration values come
    from HBEFA 4.1, from 1909 to 2020 (extrapolated to 2050).

    For the other countries, values come from
    Miller, J. D., & Jin, L. (2019). Global progress toward
    soot-free diesel vehicles in 2019.
    International Council on Clean Transportation.
    https://www.theicct.org/publications/global-progress-toward-soot-free-diesel-vehicles-2019

    There is an assumption made: countries that have
    high-sulfur content fuels (above 50 ppm in 2019) are assumed to
    improve over time to reach 50 ppm by 2050.

    :return: An axarray with 'country' and 'year' as dimensions
    :rtype: xarray.core.dataarray.DataArray
    """
    filename = "S_concentration_fuel.csv"
    filepath = DATA_DIR / "fuel" / filename

    if not filepath.is_file():
        raise FileNotFoundError(
            "The CSV file that contains sulfur concentration values could not be found."
        )
    dataframe = pd.read_csv(filepath, sep=";")
    dataframe = dataframe.groupby(["country", "year"]).sum().unstack()
    dataframe.loc[:, ("diesel", 1990)] = dataframe["diesel"].max(1)
    dataframe.loc[:, ("petrol", 1990)] = dataframe["petrol"].max(1)

    dataframe.loc[dataframe[("diesel", 2019)] > 50 / 1e6, ("diesel", 2050)] = 50 / 1e6
    dataframe.loc[dataframe[("petrol", 2019)] > 50 / 1e6, ("petrol", 2050)] = 50 / 1e6
    dataframe.loc[:, ("diesel", 2050)] = dataframe["diesel"].min(1)
    dataframe.loc[:, ("petrol", 2050)] = dataframe["petrol"].min(1)

    dataframe = dataframe.interpolate(axis=1)
    dataframe = dataframe.unstack().reset_index()
    dataframe = dataframe.rename(columns={"level_0": "fuel"})
    arr = dataframe.groupby(["country", "year", "fuel"]).sum()[0].to_xarray()
    return arr


def get_default_fuels() -> dict:
    """
    Import default fuels from `default_fuels.yaml`
    """
    filename = "default_fuels.yaml"
    filepath = DATA_DIR / "fuel" / filename
    if not filepath.is_file():
        raise FileNotFoundError(
            "The YAML file that contains default fuels could not be found."
        )
    with open(filepath, "r") as file:
        default_fuels = yaml.safe_load(file)

    return default_fuels


def get_fuels_specs() -> dict:
    """
    Import fuel specifications from `fuel_specs.yaml`
    Contains names, LHV, CO2 emission factors.
    """
    with open(DATA_DIR / "fuel" / "fuel_specs.yaml", "r", encoding="utf-8") as stream:
        fuel_specs = yaml.safe_load(stream)

    return fuel_specs


class BackgroundSystemModel:
    """
    Retrieve and build dictionaries that contain important information to model in the background system:

        * gross electricity production mixes from nearly all countries in the world, from 2015 to 2050.
        * cumulative electricity transformation/transmission/distribution losses from high voltage to medium and low voltage.
        * share of biomass-derived fuel in the total consumption of liquid fuel in the transport sector. Source: REMIND.
        * share of bioethanol, biodiesel and biomethane, for each country, for different years.
        * share of sulfur in gasoline and diesel, for different countries and years.
    """

    def __init__(self) -> None:
        self.electricity_mix = get_electricity_mix()
        self.losses = get_electricity_losses()
        self.sulfur = get_sulfur_content_in_fuel()
        self.biomethane = get_biofuel_share(DATA_DIR / "fuel" / "share_bio_cng.csv")
        self.bioethanol = get_biofuel_share(
            DATA_DIR / "fuel" / "share_bio_gasoline.csv"
        )
        self.biodiesel = get_biofuel_share(DATA_DIR / "fuel" / "share_bio_diesel.csv")
        self.default_fuels = get_default_fuels()
        self.fuel_specs = get_fuels_specs()

    def __str__(self):
        return self.__class__.__name__

    def get_share_biofuel(self, fuel: str, country: str, years: List[int]) -> np.array:
        """
        Returns average share of biodiesel according to historical IEA stats
        with an upper limit of 30% when extrapolating.
        """

        map_array = {
            "biodiesel": self.biodiesel,
            "bioethanol": self.bioethanol,
            "biomethane": self.biomethane,
        }

        share_biofuel = np.squeeze(
            np.clip(
                map_array[fuel]
                .sel(country=country)
                .interp(
                    year=years,
                    kwargs={"fill_value": "extrapolate"},
                )
                .values,
                0,
                0.3,
            )
        )
        return share_biofuel

    def find_fuel_shares(
        self, fuel_blend: dict, fuel_type: str, country: str, years: List[int]
    ) -> [str, str, np.array, np.array]:
        """
        Find the fuel shares of the fuel blend, given a country, a fuel type and a list of years.
        """
        if fuel_type in fuel_blend:
            primary = fuel_blend[fuel_type]["primary"]["type"]

            try:
                # See if a secondary fuel type has been specified
                secondary = fuel_blend[fuel_type]["secondary fuel"]["type"]

            except KeyError:
                # A secondary fuel has not been specified, set one by default
                # Check first if the default fuel is not similar to the primary fuel

                if self.default_fuels[fuel_type]["secondary"] != primary:
                    secondary = self.default_fuels[fuel_type]["secondary"]
                else:
                    secondary = [
                        f for f in self.default_fuels[fuel_type]["all"] if f != primary
                    ][0]

            secondary_share = np.array(1) - np.zeros_like(np.array(years))

        else:
            primary = self.default_fuels[fuel_type]["primary"]
            secondary = self.default_fuels[fuel_type]["secondary"]

            if primary == "electrolysis":
                secondary_share = np.zeros_like(np.array(years))

            else:
                if fuel_type == "diesel":
                    if country in self.biodiesel.country.values:
                        secondary_share = self.get_share_biofuel(
                            "biodiesel", country, years
                        )
                    else:
                        secondary_share = self.get_share_biofuel(
                            "biodiesel", "RER", years
                        )

                elif fuel_type == "cng":
                    if country in self.biomethane.country.values:
                        secondary_share = self.get_share_biofuel(
                            "biomethane", country, years
                        )
                    else:
                        secondary_share = self.get_share_biofuel(
                            "biomethane", "RER", years
                        )
                else:
                    if country in self.bioethanol.country.values:
                        secondary_share = self.get_share_biofuel(
                            "bioethanol", country, years
                        )
                    else:
                        secondary_share = self.get_share_biofuel(
                            "bioethanol", "RER", years
                        )

        primary_share = 1 - np.squeeze(np.array(secondary_share))
        secondary_share = np.squeeze(secondary_share)

        return primary, secondary, primary_share, secondary_share

    def define_fuel_blends(
        self, powertrains: List[str], country: str, years: List[int]
    ) -> dict:
        """
        This function defines fuel blends from what is passed in `fuel_blend`.
        It populates a dictionary `self.fuel_blends` that contains the
        respective shares, lower heating values
        and CO2 emission factors of the fuels used.

        Source for LHV: https://www.bafu.admin.ch/bafu/en/home/topics/climate/state/data/climate-reporting/references.html

        :return:
        """

        fuel_blend = dict()

        fuel_to_powertrains_map = {
            "diesel": ["ICEV-d", "HEV-d", "PHEV-d", "PHEV-c-d"],
            "petrol": ["ICEV-p", "HEV-p", "PHEV-p", "PHEV-c-p"],
            "cng": ["ICEV-g"],
            "hydrogen": ["FCEV"],
        }

        _arr = lambda x: np.asarray(x) if not isinstance(x, np.ndarray) else x

        for fuel_type, pwt in fuel_to_powertrains_map.items():
            if any(i in powertrains for i in pwt):
                (
                    primary,
                    secondary,
                    primary_share,
                    secondary_share,
                ) = self.find_fuel_shares(fuel_blend, fuel_type, country, years)

                primary_share = _arr(primary_share)
                secondary_share = _arr(secondary_share)

                # add a trailing dimension to the array if it is 0D
                if primary_share.ndim == 0:
                    primary_share = np.expand_dims(primary_share, axis=0)
                if secondary_share.ndim == 0:
                    secondary_share = np.expand_dims(secondary_share, axis=0)

                fuel_blend[fuel_type] = {
                    "primary": {
                        "type": primary,
                        "share": primary_share,
                        "lhv": self.fuel_specs[primary]["lhv"],
                        "CO2": self.fuel_specs[primary]["co2"],
                        "density": self.fuel_specs[primary]["density"],
                        "name": tuple(self.fuel_specs[primary]["name"]),
                        "biogenic share": self.fuel_specs[primary]["biogenic_share"],
                    },
                    "secondary": {
                        "type": secondary,
                        "share": secondary_share,
                        "lhv": self.fuel_specs[secondary]["lhv"],
                        "CO2": self.fuel_specs[secondary]["co2"],
                        "density": self.fuel_specs[secondary]["density"],
                        "name": tuple(self.fuel_specs[secondary]["name"]),
                        "biogenic share": self.fuel_specs[secondary]["biogenic_share"],
                    },
                }

        return fuel_blend



"""

Submodules
==========

.. autosummary::
    :toctree: _autosummary


"""

__all__ = (
    "get_standard_driving_cycle_and_gradient",
    "NoiseEmissionsModel",
    "HotEmissionsModel",
    "Inventory",
    "BackgroundSystemModel",
    "ExportInventory",
    "VehicleInputParameters",
)
__version__ = (1, 0, 9)

from pathlib import Path

DATA_DIR = Path(__file__).resolve().parent / "data"

from .background_systems import BackgroundSystemModel
from .driving_cycles import get_standard_driving_cycle_and_gradient
from .export import ExportInventory
from .hot_emissions import HotEmissionsModel
from .inventory import Inventory
from .noise_emissions import NoiseEmissionsModel
from .vehicle_input_parameters import VehicleInputParameters



"""
energy_consumption.py contains the class EnergyConsumption Model
which exposes two methods:
one for calculating the auxiliary energy needs,
and another one for calculating the motive
energy needs.
"""

import csv
from typing import Any, List, Tuple, Union

import numexpr as ne
import numpy as np
import pandas as pd
import xarray as xr
import yaml
from numpy import ndarray
from xarray import DataArray

from . import DATA_DIR
from .driving_cycles import (
    get_driving_cycle_specs,
    get_standard_driving_cycle_and_gradient,
)

MONTHLY_AVG_TEMP = "monthly_avg_temp.csv"


def _(obj: Union[np.ndarray, xr.DataArray]) -> Union[np.ndarray, xr.DataArray]:
    """Add a trailing dimension to make input arrays broadcast correctly"""
    if isinstance(obj, (np.ndarray, xr.DataArray)):
        return np.expand_dims(obj, -1)

    return obj


def __(obj: Union[np.ndarray, xr.DataArray]) -> Union[np.ndarray, xr.DataArray]:
    """Add a heading dimension to make input arrays broadcast correctly"""
    if isinstance(obj, (np.ndarray, xr.DataArray)):
        return np.expand_dims(obj, 0)

    return obj


def get_default_driving_cycle_name(vehicle_type) -> str:
    """Get the default driving cycle name"""
    return list(get_driving_cycle_specs()["columns"][vehicle_type].keys())[0]


def get_efficiency_coefficients(vehicle_type: str) -> [Any, None]:
    # load yaml file to retrieve efficiency coefficients

    # if file does not exist, return None
    if not (DATA_DIR / "efficiency" / f"{vehicle_type}.yaml").exists():
        return None

    with open(DATA_DIR / "efficiency" / f"{vehicle_type}.yaml") as f:
        efficiency_coefficients = yaml.load(f, Loader=yaml.FullLoader)

    return efficiency_coefficients


def get_country_temperature(country):
    """
    Retrieves mothly average temperature
    :type country: country for which to retrieve temperature values
    :return:
    """

    with open(DATA_DIR / MONTHLY_AVG_TEMP) as f:
        reader = csv.reader(f, delimiter=";")
        for row in reader:
            if row[2] == country:
                return np.array([float(i) for i in row[3:]])

    print(
        f"Could not find monthly average temperature series for {country}. "
        f"Uses those for CH instead."
    )

    with open(DATA_DIR / MONTHLY_AVG_TEMP) as f:
        reader = csv.reader(f, delimiter=";")
        for row in reader:
            if row[2] == "CH":
                return np.array([int(i) for i in row[3:]])


def convert_to_xr(data):
    return xr.DataArray(
        data,
        dims=["second", "value", "year", "powertrain", "size", "parameter"],
        coords={
            "second": range(0, data.shape[0]),
            "value": range(0, data.shape[1]),
            "year": range(0, data.shape[2]),
            "powertrain": range(0, data.shape[3]),
            "size": range(0, data.shape[4]),
            "parameter": [
                "rolling resistance",
                "air resistance",
                "gradient resistance",
                "kinetic energy",
                "motive energy at wheels",
                "motive energy",
                "negative motive energy",
                "recuperated energy",
                "auxiliary energy",
                "cooling energy",
                "heating energy",
                "battery cooling energy",
                "battery heating energy",
                "power load",
                "transmission efficiency",
                "engine efficiency",
                "velocity",
            ],
        },
    )


class EnergyConsumptionModel:
    """
    Calculate energy consumption of a vehicle for a
    given driving cycle and vehicle parameters.

    Based on a selected driving cycle, this class calculates
    the acceleration needed and provides
    two methods:

        - :func:`~energy_consumption.EnergyConsumptionModel.aux_energy_per_km` calculates the energy needed to power auxiliary services
        - :func:`~energy_consumption.EnergyConsumptionModel.motive_energy_per_km` calculates the energy needed to move the vehicle over 1 km

    Acceleration is calculated as the difference between velocity
    at t_2 and velocity at t_0, divided by 2.
    See for example: http://www.unece.org/fileadmin/DAM/trans/doc/2012/wp29grpe/WLTP-DHC-12-07e.xls

    :param cycle: Driving cycle. Pandas Series of second-by-second speeds (km/h) or name (str)
    :type cycle: np.ndarray
    :param rho_air: Mass per unit volume of air. Set to (1.225 kg/m3) by default.
    :type rho_air: float
    :param gradient: Road gradient per second of driving, in degrees.
    None by default. Should be passed as an array of length equal
    to the length of the driving cycle.
    :type gradient: numpy.ndarray

    :ivar rho_air: Mass per unit volume of air. Value of 1.204 at 23C (test temperature for WLTC).
    :vartype rho_air: float
    :ivar velocity: Time series of speed values, in meters per second.
    :vartype velocity: numpy.ndarray
    :ivar acceleration: Time series of acceleration, calculated as
    increment in velocity per interval of 1 second, in meter per second^2.
    :vartype acceleration: numpy.ndarray

    """

    def __init__(
        self,
        vehicle_type: str,
        vehicle_size: List[str],
        powertrains: List[str],
        cycle: Union[str, np.ndarray],
        gradient: Union[str, np.ndarray],
        rho_air: float = 1.204,
        country: str = "CH",
        ambient_temperature: Union[float, np.ndarray] = None,
        indoor_temperature: Union[float, np.ndarray] = 20,
    ) -> None:
        if not isinstance(vehicle_size, list):
            vehicle_size = [vehicle_size]

        self.rho_air = rho_air

        if isinstance(cycle, np.ndarray):
            self.cycle_name = "custom"
            self.cycle = cycle.reshape(-1, 1)

            if gradient is not None:
                self.gradient = gradient.reshape(-1, 1)
            else:
                self.gradient = np.zeros_like(self.cycle)

        else:
            self.cycle_name = cycle
            self.cycle, self.gradient = get_standard_driving_cycle_and_gradient(
                vehicle_type, vehicle_size, self.cycle_name
            )

        self.country = country
        self.vehicle_type = vehicle_type
        self.vehicle_size = vehicle_size
        self.powertrains = powertrains

        if "Micro" in vehicle_size:
            idx = vehicle_size.index("Micro")
            self.cycle[:, idx] = np.clip(self.cycle[:, idx], 0, 90)

        assert len(self.cycle) == len(
            self.gradient
        ), "The length of the driving cycle and the gradient must be the same."

        # Unit conversion km/h to m/s
        self.velocity = np.where(np.isnan(self.cycle), 0, (self.cycle * 1000) / 3600)
        self.velocity = self.velocity[:, None, None, None, :]
        self.driving_time = self.find_last_driving_second()

        # Model acceleration as difference in velocity between
        # time steps (1 second)
        # Zero at first value
        self.acceleration = np.zeros_like(self.velocity)
        self.acceleration[1:-1] = (self.velocity[2:, ...] - self.velocity[:-2, ...]) / 2

        self.efficiency_coefficients = get_efficiency_coefficients(vehicle_type)

        self.ambient_temperature = ambient_temperature
        self.indoor_temperature = indoor_temperature

    def calculate_hvac_energy(
        self,
        hvac_power,
        battery_cooling_unit,
        battery_heating_unit,
    ) -> tuple[Any, Any, Any, Any]:
        if self.ambient_temperature is not None:
            if isinstance(self.ambient_temperature, (float, int)):
                self.ambient_temperature = np.resize(self.ambient_temperature, (12,))
            else:
                self.ambient_temperature = np.array(self.ambient_temperature)
                assert (
                    len(self.ambient_temperature) == 12
                ), "Ambient temperature must be a 12-month array"
        else:
            self.ambient_temperature = np.resize(
                get_country_temperature(self.country), (12,)
            )

        if self.indoor_temperature is not None:
            if isinstance(self.indoor_temperature, (float, int)):
                self.indoor_temperature = np.resize(self.indoor_temperature, (12,))
            else:
                self.indoor_temperature = np.array(self.indoor_temperature)
                assert (
                    len(self.indoor_temperature) == 12
                ), "Indoor temperature must be a 12-month array"

        # use ambient temperature if provided, otherwise
        # monthly temperature average (12 values)
        # relation between ambient temperature
        # and HVAC power required
        # from https://doi.org/10.1016/j.energy.2018.12.064
        amb_temp_data_points = np.array([-30, -20, -10, 0, 10, 20, 30, 40])
        pct_power_HVAC = np.array([0.95, 0.54, 0.29, 0.13, 0.04, 0.08, 0.45, 0.7])

        # Heating power as long as ambient temperature, in W
        # is below the comfort indoor temperature
        p_heating = (
            np.where(
                self.ambient_temperature < self.indoor_temperature,
                np.interp(
                    self.ambient_temperature, amb_temp_data_points, pct_power_HVAC
                ),
                0,
            ).mean()
            * hvac_power
        ).values

        # Cooling power as long as ambient temperature, in W
        # is above the comfort indoor temperature
        p_cooling = (
            np.where(
                self.ambient_temperature >= self.indoor_temperature,
                np.interp(
                    self.ambient_temperature, amb_temp_data_points, pct_power_HVAC
                ),
                0,
            ).mean()
            * hvac_power
        ).values

        # We want to add power draw for battery cooling
        # and battery heating

        # battery cooling occurring above 20C, in W
        p_battery_cooling = np.where(
            self.ambient_temperature > 20, _(battery_cooling_unit), 0
        )
        p_battery_cooling = p_battery_cooling.mean(-1)

        # battery heating occurring below 5C, in W
        p_battery_heating = np.where(
            self.ambient_temperature < 5, _(battery_heating_unit), 0
        )
        p_battery_heating = p_battery_heating.mean(-1)

        return p_cooling, p_heating, p_battery_cooling, p_battery_heating

    def find_last_driving_second(self) -> ndarray:
        """
        Find the last second of the driving cycle that is not zero.
        """

        # find last index where velocity is greater than 0
        # along the last axis of self.velocity
        # let's iterate through the last axis of self.velocity
        # and find the last index where velocity is greater than 0
        driving_time = np.zeros_like(self.velocity)

        for i in range(self.velocity.shape[-1]):
            last_index = np.where(self.velocity[..., i] > 0)[0][-1]
            driving_time[:last_index, ..., i] = 1

        return driving_time

    def aux_energy_per_km(
        self,
        aux_power: Union[xr.DataArray, np.array],
        efficiency: Union[xr.DataArray, np.array],
        hvac_power: Union[xr.DataArray, np.array] = None,
        battery_cooling_unit: Union[xr.DataArray, np.array] = None,
        battery_heating_unit: Union[xr.DataArray, np.array] = None,
        heat_pump_cop_cooling: Union[xr.DataArray, np.array] = None,
        heat_pump_cop_heating: Union[xr.DataArray, np.array] = None,
        cooling_consumption: Union[xr.DataArray, np.array] = None,
        heating_consumption: Union[xr.DataArray, np.array] = None,
    ) -> Union[tuple[Any, Any, Any, Any, Any], Any]:
        """
        Calculate energy used other than motive energy per km driven.

        :param aux_power: Total power needed for auxiliaries, heating, and cooling (W)
        :param efficiency: Efficiency of electricity generation (dimensionless, between 0.0 and 1.0).
                Battery electric vehicles should have efficiencies of one here, as we account for
                battery efficiencies elsewhere.
        :returns: total auxiliary energy in kJ/km

        """

        _o = lambda x: np.where(x == 0, 1, x)

        if hvac_power is not None:
            (
                p_cooling,
                p_heating,
                p_battery_cooling,
                p_battery_heating,
            ) = self.calculate_hvac_energy(
                hvac_power=hvac_power,
                battery_cooling_unit=battery_cooling_unit,
                battery_heating_unit=battery_heating_unit,
            )

            return (
                aux_power.T.values * np.where(self.velocity > 0, 1, 0),
                (p_cooling / _o(heat_pump_cop_cooling) * cooling_consumption).T.values
                * self.driving_time,
                (p_heating / _o(heat_pump_cop_heating) * heating_consumption).T.values
                * self.driving_time,
                p_battery_cooling.T * self.driving_time,
                p_battery_heating.T * self.driving_time,
            )

        _c = lambda x: x.values if isinstance(x, xr.DataArray) else x

        # Provide energy in kJ / km (1 J = 1 Ws)
        auxiliary_energy = (
            _c(aux_power).T[None, ...] * 1000 / 1000  # Watts  # m / km  # 1 / (J / kJ)
        )

        efficiency = _c(efficiency)

        return auxiliary_energy / _o(efficiency)

    def calculate_efficiency(
        self,
        efficiency: xr.DataArray,
        engine_load: xr.DataArray,
        efficiency_type: str,
    ) -> xr.DataArray:
        if self.efficiency_coefficients is None:
            return xr.where(efficiency == 0, 1, efficiency)

        # Calculate efficiency based on engine load
        pwts = {
            "ICEV-p": "gasoline",
            "ICEV-d": "diesel",
            "HEV-p": "gasoline",
            "HEV-d": "diesel",
            "ICEV-g": "compressed gas",
            "PHEV-c-p": "gasoline",
            "PHEV-c-d": "diesel",
            "BEV": "electric",
            "BEV-depot": "electric",
            "BEV-opp": "electric",
            "BEV-motion": "electric",
            "PHEV-e": "electric",
            "FCEV": "electric",
        }

        # Calculate efficiency based on engine load
        for p, pwt in enumerate(self.powertrains):
            if pwt not in pwts:
                continue

            efficiency[:, :, :, p] = np.clip(
                np.interp(
                    engine_load[:, :, :, p],
                    np.fromiter(
                        self.efficiency_coefficients[pwts[pwt]][efficiency_type].keys(),
                        dtype=float,
                    ),
                    np.fromiter(
                        self.efficiency_coefficients[pwts[pwt]][
                            efficiency_type
                        ].values(),
                        dtype=float,
                    ),
                ),
                0.00,
                1.0,
            )

        return efficiency

    def motive_energy_per_km(
        self,
        driving_mass: Union[xr.DataArray, np.array],
        rr_coef: Union[xr.DataArray, np.array],
        drag_coef: Union[xr.DataArray, np.array],
        frontal_area: Union[xr.DataArray, np.array],
        electric_motor_power: Union[xr.DataArray, np.array],
        engine_power: Union[xr.DataArray, np.array],
        recuperation_efficiency: Union[xr.DataArray, np.array],
        aux_power: Union[xr.DataArray, np.array],
        battery_charge_eff: Union[xr.DataArray, np.array],
        battery_discharge_eff: Union[xr.DataArray, np.array],
        engine_efficiency: Union[xr.DataArray, np.array, None] = None,
        transmission_efficiency: Union[xr.DataArray, np.array, None] = None,
        fuel_cell_system_efficiency: Union[xr.DataArray, np.array] = None,
        hvac_power: Union[xr.DataArray, np.array] = None,
        battery_cooling_unit: Union[xr.DataArray, np.array] = None,
        battery_heating_unit: Union[xr.DataArray, np.array] = None,
        heat_pump_cop_cooling: Union[xr.DataArray, np.array] = None,
        heat_pump_cop_heating: Union[xr.DataArray, np.array] = None,
        cooling_consumption: Union[xr.DataArray, np.array] = None,
        heating_consumption: Union[xr.DataArray, np.array] = None,
    ) -> DataArray:
        """
        Calculate energy used and recuperated for a given vehicle per km driven.

        :param driving_mass: Mass of vehicle (kg)
        :param rr_coef: Rolling resistance coefficient (dimensionless, between 0.0 and 1.0)
        :param drag_coef: Aerodynamic drag coefficient (dimensionless, between 0.0 and 1.0)
        :param frontal_area: Frontal area of vehicle (m2)
        :param sizes: size classes of the vehicles
        :param electric_motor_power: Electric motor power (watts). Optional.
        :returns: net motive energy (in kJ/km)

        Power to overcome rolling resistance is calculated by:

        .. math::

            g v M C_{r}

        where :math:`g` is 9.81 (m/s2), :math:`v` is velocity (m/s), :math:`M` is mass (kg),
        and :math:`C_{r}` is the rolling resistance coefficient (dimensionless).

        Power to overcome air resistance is calculated by:

        .. math::

            \frac{1}{2} \rho_{air} v^{3} A C_{d}


        where :math:`\rho_{air}` is 1.225 (kg/m3), :math:`v` is velocity (m/s), :math:`A` is frontal area (m2), and :math:`C_{d}`
        is the aerodynamic drag coefficient (dimensionless).

        """

        _c = lambda x: x.values if isinstance(x, xr.DataArray) else x
        _o = lambda x: np.where((x == 0) | (x == np.nan), 1, x)

        # Calculate the energy used for each second of the drive cycle
        ones = np.ones_like(self.velocity)

        # Resistance from the tire rolling: rolling resistance coefficient * driving mass * 9.81
        rolling_resistance = _c((driving_mass * rr_coef * 9.81).T) * (self.velocity > 0)

        # Resistance from the drag: frontal area * drag coefficient * air density * 1/2 * velocity^2
        air_resistance = _c((frontal_area * drag_coef * self.rho_air / 2).T) * np.power(
            self.velocity, 2
        )

        # Resistance from road gradient: driving mass * 9.81 * sin(gradient)
        gradient_resistance = (
            _c((driving_mass * 9.81).T)
            * np.sin(np.nan_to_num(self.gradient)[:, None, None, None, :])
            * (self.velocity > 0)
        )

        # Inertia: driving mass * acceleration
        inertia = self.acceleration * _c(driving_mass).T

        total_resistance = (
            rolling_resistance + air_resistance + gradient_resistance + inertia
        )
        motive_energy_at_wheels = xr.where(total_resistance < 0, 0, total_resistance)
        motive_energy = np.zeros_like(motive_energy_at_wheels)

        # determining efficiencies
        engine_load = np.ones_like(motive_energy_at_wheels)

        if engine_efficiency is None:
            engine_efficiency = np.ones_like(motive_energy_at_wheels)

        if transmission_efficiency is None:
            transmission_efficiency = np.ones_like(motive_energy_at_wheels)

        engine_load_iterations = [0, engine_load.mean()]

        # we loop while the last three iterations are roughly equal
        # or while len(engine_load_iterations) < 10

        while len(engine_load_iterations) < 10:
            engine_efficiency = self.calculate_efficiency(
                engine_efficiency, engine_load, "engine"
            )

            transmission_efficiency = self.calculate_efficiency(
                transmission_efficiency, engine_load, "transmission"
            )

            recuperation_efficiency = xr.where(
                recuperation_efficiency == 0, 1, recuperation_efficiency
            )

            if fuel_cell_system_efficiency is None:
                fuel_cell_system_efficiency = np.array([1.0])

            fuel_cell_system_efficiency = xr.where(
                fuel_cell_system_efficiency == 0, 1, fuel_cell_system_efficiency
            )

            motive_energy = motive_energy_at_wheels / (
                _o(_c(engine_efficiency))
                * _o(_c(transmission_efficiency))
                * _o(_c(fuel_cell_system_efficiency)).T[None, ...]
            )

            engine_load = np.clip(
                (motive_energy / (_o(_c(engine_power)).T * 1000)) * self.velocity, 0, 1
            )

            # add a minimum 5% engine load when the vehicle is idling
            engine_load = np.where(self.velocity == 0, 0.05, engine_load)
            engine_load *= self.driving_time
            engine_load_iterations.append(engine_load.mean())

        negative_motive_energy = xr.where(total_resistance > 0, 0, total_resistance)
        recuperated_energy = (
            negative_motive_energy
            * _c(recuperation_efficiency).T[None, ...]
            * _c(battery_charge_eff).T[None, ...]
            * _c(battery_discharge_eff).T[None, ...]
            * (_c(electric_motor_power).T[None, ...] > 0)
        )

        if hvac_power is None:
            auxiliary_energy = self.aux_energy_per_km(
                aux_power,
                engine_efficiency,
                hvac_power,
                battery_cooling_unit,
                battery_heating_unit,
                heat_pump_cop_cooling,
                heat_pump_cop_heating,
                cooling_consumption,
                heating_consumption,
            )
            cooling_energy, heating_energy, battery_cooling, battery_heating = (
                np.zeros_like(auxiliary_energy),
                np.zeros_like(auxiliary_energy),
                np.zeros_like(auxiliary_energy),
                np.zeros_like(auxiliary_energy),
            )
        else:
            (
                auxiliary_energy,
                cooling_energy,
                heating_energy,
                battery_cooling,
                battery_heating,
            ) = self.aux_energy_per_km(
                aux_power,
                engine_efficiency,
                hvac_power,
                battery_cooling_unit,
                battery_heating_unit,
                heat_pump_cop_cooling,
                heat_pump_cop_heating,
                cooling_consumption,
                heating_consumption,
            )

        auxiliary_energy *= self.velocity > 0

        all_arrays = np.concatenate(
            [
                _(rolling_resistance),
                _(air_resistance),
                _(gradient_resistance),
                _(inertia),
                _(motive_energy_at_wheels),
                _(motive_energy),
                _(negative_motive_energy),
                _(recuperated_energy),
                _(auxiliary_energy),
                _(cooling_energy),
                _(heating_energy),
                _(battery_cooling),
                _(battery_heating),
                _(engine_load),
                _(transmission_efficiency),
                _(engine_efficiency),
                _(self.velocity * np.ones_like(motive_energy)),
            ],
            axis=-1,
        )

        all_arrays[..., :-4] /= 1000
        all_arrays[..., :-9] *= _(self.velocity)

        all_arrays[..., 5] = np.where(
            all_arrays[..., 5] > _(engine_power).T * 1,
            _(engine_power).T * 1,
            all_arrays[..., 5],
        )
        all_arrays[..., 7] = np.where(
            all_arrays[..., 7] < _(electric_motor_power).T * -1,
            _(electric_motor_power).T * -1,
            all_arrays[..., 7],
        )

        return convert_to_xr(all_arrays).fillna(0)



"""
particulates_emissions.py contains ParticulatesEmissionsModel which calculated the amount of
abrasion particles emitted, given a driving cycle.
"""
from typing import Union

import numpy as np
import xarray as xr

from .hot_emissions import get_driving_cycle_compartments


def _(obj: Union[np.ndarray, xr.DataArray]) -> Union[np.ndarray, xr.DataArray]:
    """Add a trailing dimension to make input arrays broadcast correctly"""
    if isinstance(obj, (np.ndarray, xr.DataArray)):
        return np.expand_dims(obj, -1)
    return obj


class ParticulatesEmissionsModel:
    """
    Calculate particulates emissions based on the method described in:
    https://www.eea.europa.eu/ds_resolveuid/6USNA27I4D

    and further disaggregated in:
    https://doi.org/10.1016/j.atmosenv.2020.117886

    Include emission from:

    - brake wear
    - tire wear
    - road wear
    - re-suspended road dust

    by considering:

    - vehicle mass
    - driving situation (urban, rural, motorway)

    into the following fractions:

    - PM 2.5
    - PM 10

    Emissions are subdivided in compartments: urban, suburban and rural.

    :param velocity: Driving cycle. Pandas Series of second-by-second speeds (km/h) or name (str)
        of cycle e.g., "Urban delivery", "Regional delivery", "Long haul".
    :param cycle_name: name of the driving cycle. Str.


    """

    def __init__(self, velocity: xr.DataArray, mass: xr.DataArray) -> None:
        self.mass = mass.values / 1000  # in tons
        self.velocity = velocity / 1000 * 3600  # in km/h
        self.distance = velocity.sum(dim="second") / 1000

    def get_abrasion_emissions(self) -> np.ndarray:
        (
            tire_pm10_urban,
            tire_pm10_rural,
            tire_pm10_motorway,
            tire_pm25_urban,
            tire_pm25_rural,
            tire_pm25_motorway,
        ) = self.get_tire_wear_emissions()

        (
            brake_pm10_urban,
            brake_pm10_rural,
            brake_pm10_motorway,
            brake_pm25_urban,
            brake_pm25_rural,
            brake_pm25_motorway,
        ) = self.get_brake_wear_emissions()

        road_pm10, road_pm25 = self.get_road_wear_emissions()
        dust_pm10, dust_pm25 = self.get_resuspended_road_dust()

        urban_share = self.velocity.where(self.velocity < 50, 0).sum(
            dim="second"
        ) / self.velocity.sum(dim="second")
        suburban_share = self.velocity.where(
            (self.velocity > 50) & (self.velocity <= 80), 0
        ).sum(dim="second") / self.velocity.sum(dim="second")
        rural_share = self.velocity.where(self.velocity > 80, 0).sum(
            dim="second"
        ) / self.velocity.sum(dim="second")

        tire_wear = (tire_pm10_urban + tire_pm25_urban) * urban_share.T
        tire_wear += (tire_pm10_rural + tire_pm25_rural) * suburban_share.T
        tire_wear += (tire_pm10_motorway + tire_pm25_motorway) * rural_share.T

        brake_wear = (brake_pm10_urban + brake_pm25_urban) * urban_share.T
        brake_wear += (brake_pm10_rural + brake_pm25_rural) * suburban_share.T
        brake_wear += (brake_pm10_motorway + brake_pm25_motorway) * rural_share.T

        road_wear = road_pm10 + road_pm25
        road_dust = dust_pm10 + dust_pm25

        res = np.concatenate(
            [
                _(tire_wear),
                _(brake_wear),
                _(road_wear),
                _(road_dust),
            ],
            axis=-1,
        )

        return res.transpose(0, 1, 4, 2, 3)

    def get_tire_wear_emissions(self):
        """
        Returns tire wear emissions.

        :return:
        """

        # converted to kg per vkm
        pm10_urban = (5.8 * self.mass ** (1 / 2.3)) / 1e6
        pm25_urban = (8.2 * self.mass ** (1 / 2.3)) / 1e6

        pm10_rural = (4.5 * self.mass ** (1 / 2.3)) / 1e6
        pm25_rural = (6.4 * self.mass ** (1 / 2.3)) / 1e6

        pm10_motorway = (3.8 * self.mass ** (1 / 2.3)) / 1e6
        pm25_motorway = (5.5 * self.mass ** (1 / 2.3)) / 1e6

        return (
            pm10_urban,
            pm10_rural,
            pm10_motorway,
            pm25_urban,
            pm25_rural,
            pm25_motorway,
        )

    def get_brake_wear_emissions(self):
        """
        Returns brake wear emissions.

        :return:
        """
        # converted to kg per vkm
        pm10_urban = (4.2 * np.power(self.mass, 1 / 1.9)) / 1e6
        pm25_urban = (11 * np.power(self.mass, 1 / 1.9)) / 1e6

        pm10_rural = (1.8 * np.power(self.mass, 1 / 1.5)) / 1e6
        pm25_rural = (4.5 * np.power(self.mass, 1 / 1.5)) / 1e6

        pm10_motorway = (0.4 * np.power(self.mass, 1 / 1.3)) / 1e6
        pm25_motorway = (1.0 * np.power(self.mass, 1 / 1.3)) / 1e6

        return (
            pm10_urban,
            pm10_rural,
            pm10_motorway,
            pm25_urban,
            pm25_rural,
            pm25_motorway,
        )

    def get_road_wear_emissions(self):
        """
        Returns road wear emissions.

        :return:
        """
        # converted to kg per vkm
        pm10 = (2.8 * np.power(self.mass, 1 / 1.5)) / 1e6
        pm25 = (5.1 * np.power(self.mass, 1 / 1.5)) / 1e6

        return (pm10, pm25)

    def get_resuspended_road_dust(self):
        """
        Returns re-suspended road dust emissions.

        :return:
        """
        # converted to kg per vkm
        pm10 = (2 * np.power(self.mass, 1 / 1.1)) / 1e6
        pm25 = (8.2 * np.power(self.mass, 1 / 1.1)) / 1e6

        return (pm10, pm25)



from itertools import product
from pathlib import Path
from typing import Dict, List, Union

import numexpr as ne
import numpy as np
import xarray as xr
import yaml

from .background_systems import BackgroundSystemModel
from .driving_cycles import detect_vehicle_type
from .energy_consumption import get_default_driving_cycle_name
from .hot_emissions import HotEmissionsModel
from .noise_emissions import NoiseEmissionsModel
from .particulates_emissions import ParticulatesEmissionsModel


def finite(array, mask_value=0):
    return np.where(np.isfinite(array), array, mask_value)


class VehicleModel:

    """
    This class represents the entirety of the vehicles considered,
    with useful attributes, such as an array that stores
    all the vehicles parameters.

    :ivar array: multi-dimensional numpy-like array that contains parameters' value(s)
    :ivar cycle: name of a driving cycle, or custom driving cycle
    :ivar gradient: series of gradients, for each second of the driving cycle
    :ivar energy_storage: dictionary with selection of battery chemistry for each powertrain

    """

    DATA_DIR = Path(__file__).resolve().parent / "data"

    def __init__(
        self,
        array: xr.DataArray,
        country="CH",
        cycle: Union[None, str, np.ndarray] = None,
        gradient: Union[None, np.ndarray] = None,
        energy_storage: Union[None, Dict] = None,
        electric_utility_factor: float = None,
        drop_hybrids: bool = True,
        payload=None,
        annual_mileage=None,
        energy_target=None,
        energy_consumption: dict = None,
        engine_efficiency: dict = None,
        transmission_efficiency: dict = None,
        target_range: dict = None,
        target_mass: dict = None,
        power: dict = None,
        fuel_blend: dict = None,
        ambient_temperature: float = None,
        indoor_temperature: float = 20,
    ) -> None:
        """
        :param array: multi-dimensional numpy-like array that contains parameters' value(s)
        :param country: country code
        :param cycle: name of a driving cycle, or custom driving cycle
        :param gradient: series of gradients, for each second of the driving cycle
        :param energy_storage: dictionary with selection of battery chemistry, capacity and origin for each powertrain-size-year combination
        :param electric_utility_factor: fraction of electricity that is generated from renewable sources
        :param drop_hybrids: boolean, if True, hybrid vehicles are dropped from the inventory
        :param payload: dictionary with payload for each powertrain-size-year combination
        :param energy_target: dictionary with energy target for each year
        :param energy_consumption: dictionary with energy consumption for each powertrain-size-year combination
        :param target_range: dictionary with target range for each powertrain-size-year combination

        """
        self.array = array
        self.country = country

        self.vehicle_type = detect_vehicle_type(list(self.array.coords["size"].values))
        self.cycle = (
            cycle
            if type(cycle) in [np.ndarray, str, list]
            else get_default_driving_cycle_name(self.vehicle_type)
        )

        self.gradient = gradient
        self.energy_storage = energy_storage or {}
        self.energy_target = energy_target or {2025: 0.85, 2030: 0.7, 2050: 0.6}
        self.payload = payload or {}
        self.annual_mileage = annual_mileage or {}

        self.set_battery_chemistry()
        self.set_battery_preferences()
        self.energy = None
        self.electric_utility_factor = electric_utility_factor
        self.drop_hybrids = drop_hybrids
        self.energy_consumption = energy_consumption or None
        self.engine_efficiency = engine_efficiency or None
        self.transmission_efficiency = transmission_efficiency or None

        # a range to reach can be defined by the user
        self.target_range = target_range
        # a curb mass to reach can be defined by the user
        self.target_mass = target_mass
        # overrides the engine/motor power
        self.power = power

        self.bs = BackgroundSystemModel()

        if fuel_blend:
            self.fuel_blend = self.check_fuel_blend(fuel_blend)
        else:
            self.fuel_blend = self.bs.define_fuel_blends(
                self.array.powertrain.values, self.country, self.array.year.values
            )

        self.ambient_temperature = ambient_temperature
        self.indoor_temperature = indoor_temperature

    def __call__(self, key: Union[str, List]):
        """
        This method fixes a dimension of the `array` attribute given
        a powertrain technology selected.
        Set up this class as a context manager,
        so we can have some nice syntax

        .. code-block:: python

            with class('some powertrain') as cpm:
                cpm['something']. # Will be filtered for the correct powertrain

        On with block exit, this filter is cleared
        https://stackoverflow.com/a/10252925/164864

        :param key: A powertrain type, e.g., "FCEV"
        :return: An instance of `array` filtered after the powertrain selected.

        """
        if isinstance(key, str):
            key = [key]

        self.__cache = self.array
        self.array = self.array.loc[
            dict(powertrain=[k for k in key if k in self.array.powertrain])
        ]
        return self

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.array = self.__cache
        del self.__cache

    def __getitem__(self, key: Union[str, List]) -> xr.DataArray:
        """
        Make class['foo'] automatically filter for the parameter 'foo'
        Makes the model code much cleaner

        :param key: Parameter name
        :return: `array` filtered after the parameter selected
        """

        return self.array.loc[dict(parameter=key)]

    def __setitem__(self, key, value):
        self.array.loc[{"parameter": key}] = value

    def set_all(self):
        pass

    def set_battery_chemistry(self):
        pass

    def set_battery_preferences(self):
        l_parameters = [
            p
            for p in [
                "battery cell energy density",
                "battery cell mass share",
                "battery cycle life",
            ]
            if p in self.array.parameter.values
        ]

        for key, val in self.energy_storage["electric"].items():
            pwt, size, year = key

            if (
                (val is not None)
                & (pwt in self.array.powertrain.values)
                & (year in self.array.year.values)
                & (size in self.array["size"].values)
            ):
                cell_params = self.array.loc[
                    dict(
                        powertrain=pwt,
                        size=size,
                        year=year,
                        parameter=[
                            f"{p}, {val.split('-')[0].strip()}" for p in l_parameters
                        ],
                    )
                ]

                self.array.loc[
                    dict(
                        powertrain=pwt,
                        size=size,
                        year=year,
                        parameter=l_parameters,
                    )
                ] = cell_params.values

    def adjust_cost(self) -> None:
        """
        This method adjusts costs of energy storage over time, to correct for the overly optimistic linear
        interpolation between years.

        """

        n_iterations = self.array.shape[-1]
        n_year = len(self.array.year.values)

        # If uncertainty is not considered, the cost factor equals 1.
        # Otherwise, a variability of +/-30% is added.

        if n_iterations == 1:
            cost_factor = 1
        else:
            if "reference" in self.array.value.values.tolist():
                cost_factor = np.ones((n_iterations, 1))
            else:
                cost_factor = np.random.triangular(0.7, 1, 1.3, (n_iterations, 1))

        # Correction of hydrogen tank cost, per kg
        # Correction of fuel cell stack cost, per kW
        if "FCEV" in self.array.powertrain:
            self.array.loc[
                dict(powertrain="FCEV", parameter="fuel tank cost per kg")
            ] = np.reshape(
                (1.078e58 * np.exp(-6.32e-2 * self.array.year.values) + 3.43e2)
                * cost_factor,
                (1, n_year, n_iterations),
            )

            self.array.loc[
                dict(powertrain="FCEV", parameter="fuel tank cost per kg")
            ] = np.reshape(
                (3.15e66 * np.exp(-7.35e-2 * self.array.year.values) + 2.39e1)
                * cost_factor,
                (1, n_year, n_iterations),
            )

        # Correction of energy battery system cost, per kWh
        list_batt = [
            i
            for i in ["BEV", "PHEV-e", "PHEV-c-p", "PHEV-c-d"]
            if i in self.array.powertrain
        ]
        if len(list_batt) > 0:
            self.array.loc[
                dict(powertrain=list_batt, parameter="energy battery cost per kWh")
            ] = np.reshape(
                (2.75e86 * np.exp(-9.61e-2 * self.array.year.values) + 5.059e1)
                * cost_factor,
                (1, 1, n_year, n_iterations),
            )

        # Correction of power battery system cost, per kW
        list_pwt = [
            i
            for i in [
                "ICEV-p",
                "ICEV-d",
                "ICEV-g",
                "PHEV-c-p",
                "PHEV-c-d",
                "FCEV",
                "HEV-p",
                "HEV-d",
            ]
            if i in self.array.powertrain
        ]

        if len(list_pwt) > 0:
            self.array.loc[
                dict(powertrain=list_pwt, parameter="power battery cost per kW")
            ] = np.reshape(
                (8.337e40 * np.exp(-4.49e-2 * self.array.year.values) + 11.17)
                * cost_factor,
                (1, 1, n_year, n_iterations),
            )

        # Correction of combustion powertrain cost for ICEV-g
        if "ICEV-g" in self.array.powertrain:
            self.array.loc[
                dict(powertrain="ICEV-g", parameter="combustion powertrain cost per kW")
            ] = np.clip(
                np.reshape(
                    (5.92e160 * np.exp(-0.1819 * self.array.year.values) + 26.76)
                    * cost_factor,
                    (1, n_year, n_iterations),
                ),
                None,
                100,
            )

    def drop_hybrid(self) -> None:
        """
        This method drops the powertrains `PHEV-c-p`, `PHEV-c-d` and `PHEV-e` as they were only used to create the
        `PHEV` powertrain.
        :returns: Does not return anything. Modifies ``self.array`` in place.
        """
        self.array = self.array.sel(
            powertrain=[
                pt
                for pt in self.array.coords["powertrain"].values
                if pt not in ["PHEV-e", "PHEV-c-p", "PHEV-c-d"]
            ]
        )

    def set_electricity_consumption(self) -> None:
        """
        This method calculates the total electricity consumption for BEV
        and plugin-hybrid vehicles
        :returns: Does not return anything. Modifies ``self.array`` in place.
        """
        _ = lambda x: np.where(x == 0, 1, x)

        self["electricity consumption"] = (
            self["TtW energy"]
            / _(self["battery charge efficiency"])
            / _(self["charger efficiency"])
            / 3600
            * (self["charger mass"] > 0)
        )

        var = (
            "range"
            if "range" in self.array.coords["parameter"].values
            else (
                "target range"
                if "target range" in self.array.coords["parameter"].values
                else "daily distance"
            )
        )

        self["fuel consumption"] = (
            self["fuel mass"] / _(self[var]) / _(self["fuel density per kg"])
        )

    def override_ttw_energy(self):
        # override of TtW energy, provided by the user
        if self.energy_consumption:
            for key, val in self.energy_consumption.items():
                pwt, size, year = key
                if val is not None:
                    print(
                        f"Overriding TtW energy for {pwt} {size} {year} "
                        f"with {val} kj/km"
                    )

                    distance = (
                        self.energy.sel(
                            parameter="velocity",
                            powertrain=pwt,
                            size=size,
                            year=year,
                        ).sum(dim="second")
                        / 1000
                    )

                    self.energy.loc[
                        dict(
                            powertrain=pwt,
                            size=size,
                            year=year,
                            parameter="motive energy",
                        )
                    ] = (
                        val * distance / self.energy.shape[0]  # kj/km  # km  # seconds
                    )

                    self.energy.loc[
                        dict(
                            powertrain=pwt,
                            size=size,
                            year=year,
                            parameter=["auxiliary energy", "recuperated energy"],
                        )
                    ] = 0

                    # update self["TtW energy"]
                    self["TtW energy"] = (
                        self.energy.sel(
                            parameter=[
                                "motive energy",
                                "auxiliary energy",
                                "recuperated energy",
                            ]
                        ).sum(dim=["second", "parameter"])
                        / distance
                    ).T

    def calculate_ttw_energy(self) -> None:
        """
        This method calculates the energy required to operate auxiliary
        services as well as to move the car. The sum is stored under the
        parameter label "TtW energy" in :attr:`self.array`.

        """

        pass

    def set_fuel_cell_mass(self):
        """
        Specific setup for fuel cells, which are mild hybrids.
        Must be called after :meth:`.set_power_parameters`.
        """

        # our basic fuel cell mass is based
        # on a car fuel cell with 800 mW/cm2
        # the cell power density is adapted for truck or bus use
        # it is decreased comparatively to that of a passenger car
        # to reflect increased durability

        _ = lambda x: np.where(x == 0, 1, x)

        self["fuel cell stack mass"] = (
            self["fuel cell power density"]
            * self["fuel cell power"]
            * (800 / _(self["fuel cell power area density"]))
        )
        self["fuel cell ancillary BoP mass"] = (
            self["fuel cell power"] * self["fuel cell ancillary BoP mass per power"]
        )
        self["fuel cell essential BoP mass"] = (
            self["fuel cell power"] * self["fuel cell essential BoP mass per power"]
        )

        if "FCEV" in self.array.powertrain.values:
            self.array.loc[
                dict(parameter="battery power", powertrain="FCEV")
            ] = self.array.loc[dict(parameter="fuel cell power", powertrain="FCEV")] * (
                np.array(1)
                - self.array.loc[
                    dict(parameter="fuel cell power share", powertrain="FCEV")
                ]
            )

            self.array.loc[dict(parameter="battery cell mass", powertrain="FCEV")] = (
                self.array.loc[dict(parameter="battery power", powertrain="FCEV")]
                / self.array.loc[
                    dict(parameter="battery cell power density", powertrain="FCEV")
                ]
            )

            self.array.loc[
                dict(parameter="battery BoP mass", powertrain="FCEV")
            ] = self.array.loc[
                dict(parameter="battery cell mass", powertrain="FCEV")
            ] * (
                np.array(1)
                - self.array.loc[
                    dict(parameter="battery cell mass share", powertrain="FCEV")
                ]
            )

    def set_fuel_cell_power(self) -> None:
        """
        Specific setup for fuel cells, which are mild hybrids.
        Must be called after :meth:`.set_power_parameters`.
        """

        _ = lambda x: np.where(x == 0, 1, x)

        self["fuel cell system efficiency"] = (
            self["fuel cell stack efficiency"]
            / _(self["fuel cell own consumption"])
            * (self["fuel cell own consumption"] > 0)
        )

        self["fuel cell power"] = (
            self["power"]
            * self["fuel cell power share"]
            * self["fuel cell own consumption"]
        )

    def set_auxiliaries(self) -> None:
        """
        Calculates the power needed to operate the auxiliary services
        of the vehicle (heating, cooling).

        The demand for heat and cold are expressed as a fraction of the
        heating and cooling capacities

        .. note:

            Auxiliary power demand (W) = Base auxiliary power (W) +
            (Heating demand (dimensionless, between 0 and 1) * Heating power (W)) +
            (Cooling demand (dimensionless, between 0 and 1) * Cooling power (W))

        """
        self["auxiliary power demand"] = (
            self["auxilliary power base demand"]
            + self["heating thermal demand"] * self["heating energy consumption"]
            + self["cooling thermal demand"] * self["cooling energy consumption"]
        )

    def set_recuperation(self):
        _ = lambda x: np.where(x == 0, 1, x)
        self["recuperation efficiency"] = _(
            self["transmission efficiency"] * (self["combustion power share"] < 1)
        )

    def set_battery_fuel_cell_replacements(self) -> None:
        """
        Calculates the fraction of the replacement battery
        needed to match the vehicle lifetime.

        .. note::
            if ``car lifetime`` = 200000 (km) and
            ``battery lifetime`` = 190000 (km)
            then ``replacement battery`` = 0.05

        .. note::
            It is debatable whether this is realistic or not.
            Car owners may not decide to invest in a new
            battery if the remaining lifetime of the car is
            only 10000 km. Also, a battery lifetime may be expressed
            in other terms, e.g., charging cycles.

        """
        # Here we assume that we can use fractions of a battery
        # (averaged across the fleet)
        self["battery lifetime replacements"] = finite(
            np.clip(
                (self["lifetime kilometers"] / self["battery lifetime kilometers"]) - 1,
                0,
                None,
            )
        )

        # The number of fuel cell replacements is based on the
        # average distance driven with a set of fuel cells given
        # their lifetime expressed in hours of use.
        # The number of replacement is rounded *up* as we assume
        # no allocation of burden with a second life

        average_speed = (
            np.nanmean(
                np.where(
                    self.energy.sel(parameter="velocity") > 0,
                    self.energy.sel(parameter="velocity"),
                    np.nan,
                ),
                0,
            )
            * 3.6
        )

        _ = lambda array: np.where(array == 0, 1, array)

        self["fuel cell lifetime replacements"] = np.ceil(
            np.clip(
                self["lifetime kilometers"]
                / (average_speed.T * _(self["fuel cell lifetime hours"]))
                - 1,
                0,
                5,
            )
        ) * (self["fuel cell lifetime hours"] > 0)

    def override_vehicle_mass(self):
        for key, target_mass in self.target_mass.items():
            pwt, size, year = key

            if target_mass:
                current_curb_mass = self.array.loc[
                    dict(powertrain=pwt, size=size, year=year, parameter="curb mass")
                ]
                mass_difference = target_mass - current_curb_mass

                self.array.loc[
                    dict(
                        powertrain=pwt,
                        size=size,
                        year=year,
                        parameter="glider base mass",
                    )
                ] += mass_difference / (
                    1
                    - self.array.loc[
                        dict(
                            powertrain=pwt,
                            size=size,
                            year=year,
                            parameter="lightweighting",
                        )
                    ]
                )

                self.array.loc[
                    dict(powertrain=pwt, size=size, year=year, parameter="curb mass")
                ] = target_mass

    def set_vehicle_masses(self) -> None:
        """
        Define ``curb mass``, ``driving mass``, and ``total cargo mass``.

            * `curb mass <https://en.wikipedia.org/wiki/Curb_weight>`__
            is the mass of the vehicle and fuel, without people or cargo.
            * ``total cargo mass`` is the mass of the cargo and passengers.
            * ``driving mass`` is the ``curb mass`` plus ``total cargo mass``.

        .. note::
            driving mass = total cargo mass + driving mass

        """

        pass

    def override_power(self):
        if self.power:
            for key, power in self.power.items():
                pwt, size, year = key
                if power:
                    self.array.loc[
                        dict(powertrain=pwt, size=size, year=year, parameter="power")
                    ] = power

    def set_power_parameters(self) -> None:
        """
        Set electric and combustion motor powers
        based on input parameter ``power to mass ratio``.
        """
        # Convert from W/kg to kW
        self["power"] = self["power to mass ratio"] * self["curb mass"] / 1000

        if self.power:
            self.override_power()

        self["combustion power share"] = self["combustion power share"].clip(
            min=0, max=1
        )
        self["combustion power"] = self["power"] * self["combustion power share"]
        self["electric power"] = self["power"] * (
            np.array(1) - self["combustion power share"]
        )

    def set_component_masses(self) -> None:
        self["combustion engine mass"] = (
            self["combustion power"] * self["combustion mass per power"]
            + self["combustion fixed mass"]
        )

        self["electric engine mass"] = (
            self["electric power"] * self["electric mass per power"]
            + self["electric fixed mass"]
        ) * (self["electric power"] > 0)

        self["powertrain mass"] = (
            self["power"] * self["powertrain mass per power"]
            + self["powertrain fixed mass"]
        )

    def set_share_recuperated_energy(self) -> None:
        """
        Calculate the share of recuperated energy,
        over the total negative motive energy.
        """

        _ = lambda x: np.where(x == 0, 1, x)

        self["share recuperated energy"] = (
            self.energy.sel(parameter="recuperated energy").sum(dim="second")
            / _(self.energy.sel(parameter="negative motive energy").sum(dim="second"))
        ).values.T
        self["share recuperated energy"] *= self["combustion power share"] < 1

        if "PHEV-d" in self.array.powertrain:
            self.array.loc[
                dict(powertrain="PHEV-c-d", parameter="share recuperated energy")
            ] = self.array.loc[
                dict(powertrain="PHEV-e", parameter="share recuperated energy")
            ]

        if "PHEV-p" in self.array.powertrain:
            self.array.loc[
                dict(powertrain="PHEV-c-p", parameter="share recuperated energy")
            ] = self.array.loc[
                dict(powertrain="PHEV-e", parameter="share recuperated energy")
            ]

    def set_electric_utility_factor(self) -> None:
        pass

    def create_PHEV(self):
        """
        Function to create plugin-hybrid vehicles.
        PHEV-p/d is the range-weighted average
        between PHEV-c-p/PHEV-c-d and PHEV-e.
        """
        _ = lambda array: np.where(array == 0, 1, array)

        for pwt, pwtc in (("PHEV-d", "PHEV-c-d"), ("PHEV-p", "PHEV-c-p")):
            if pwt in self.array.coords["powertrain"].values:
                self.array.loc[:, pwt] = (
                    self.array.loc[:, "PHEV-e"]
                    * self.array.loc[:, "PHEV-e", "electric utility factor"]
                ) + (
                    self.array.loc[:, pwtc]
                    * (
                        np.array(1)
                        - self.array.loc[:, "PHEV-e", "electric utility factor"]
                    )
                )

                self.array.loc[:, pwt, "electric utility factor"] = self.array.loc[
                    :, "PHEV-e", "electric utility factor"
                ]

                self.energy.loc[
                    dict(
                        powertrain=pwt,
                    )
                ] = self.energy.loc[dict(powertrain="PHEV-e")]

                self.energy.loc[
                    dict(
                        powertrain=pwt,
                    )
                ] *= self.array.loc[
                    dict(parameter="electric utility factor", powertrain="PHEV-e")
                ].T.values[None, ..., None]

                self.energy.loc[
                    dict(
                        powertrain=pwt,
                    )
                ] += (
                    np.array(1)
                    - self.array.loc[
                        dict(parameter="electric utility factor", powertrain="PHEV-e")
                    ].T.values[None, ..., None]
                ) * self.energy.loc[dict(powertrain=pwtc)]

                # We need to preserve the fuel mass and fuel tank mass
                self.array.loc[
                    dict(
                        parameter=[
                            # "fuel mass",
                            "fuel tank mass",
                            "oxidation energy stored",
                            "LHV fuel MJ per kg",
                            "fuel density per kg",
                        ],
                        powertrain=pwt,
                    )
                ] = self.array.loc[
                    dict(
                        parameter=[
                            "fuel tank mass",
                            "oxidation energy stored",
                            "LHV fuel MJ per kg",
                            "fuel density per kg",
                        ],
                        powertrain=pwtc,
                    )
                ]

                # we need to preserve the battery mass from PHEV-e as well
                self.array.loc[
                    dict(
                        parameter=[
                            "energy battery mass",
                            "battery BoP mass",
                            "battery cell mass",
                            "battery DoD",
                            "battery cell energy density",
                            "battery charge efficiency",
                            "battery discharge efficiency",
                            "battery lifetime kilometers",
                            "charger efficiency",
                            "recuperation efficiency",
                            "charger mass",
                            "inverter mass",
                            "power distribution unit mass",
                        ],
                        powertrain=pwt,
                    )
                ] = self.array.loc[
                    dict(
                        parameter=[
                            "energy battery mass",
                            "battery BoP mass",
                            "battery cell mass",
                            "battery DoD",
                            "battery cell energy density",
                            "battery charge efficiency",
                            "battery discharge efficiency",
                            "battery lifetime kilometers",
                            "charger efficiency",
                            "recuperation efficiency",
                            "charger mass",
                            "inverter mass",
                            "power distribution unit mass",
                        ],
                        powertrain="PHEV-e",
                    )
                ]

                # We store the tank-to-wheel energy consumption
                # in combustion and electric mode separately
                self.array.loc[
                    dict(parameter="TtW energy", powertrain=pwt)
                ] = self.array.loc[dict(parameter="TtW energy", powertrain=pwtc)] * (
                    1
                    - self.array.loc[
                        dict(parameter="electric utility factor", powertrain="PHEV-e")
                    ]
                )
                self.array.loc[
                    dict(parameter="TtW energy", powertrain=pwt)
                ] += self.array.loc[
                    dict(parameter="TtW energy", powertrain="PHEV-e")
                ] * (
                    self.array.loc[
                        dict(parameter="electric utility factor", powertrain="PHEV-e")
                    ]
                )

                self.array.loc[
                    dict(parameter="TtW energy, combustion mode", powertrain=pwt)
                ] = self.array.loc[dict(parameter="TtW energy", powertrain=pwtc)]

                self.array.loc[
                    dict(parameter="TtW energy, electric mode", powertrain=pwt)
                ] = self.array.loc[dict(parameter="TtW energy", powertrain="PHEV-e")]

                # We need to recalculate the range as well

                var = (
                    "range"
                    if "range" in self.array.coords["parameter"].values
                    else (
                        "target range"
                        if "target range" in self.array.coords["parameter"].values
                        else "daily distance"
                    )
                )

                self.array.loc[dict(parameter=var, powertrain=pwt)] = (
                    self.array.loc[
                        dict(parameter="oxidation energy stored", powertrain=pwt)
                    ]
                    * 3600
                    / self.array.loc[
                        dict(parameter="TtW energy, combustion mode", powertrain=pwt)
                    ]
                )

                self.array.loc[dict(parameter=var, powertrain=pwt)] += (
                    self.array.loc[
                        dict(parameter="electric energy stored", powertrain=pwt)
                    ]
                    * 3600
                    / self.array.loc[
                        dict(parameter="TtW energy, electric mode", powertrain="PHEV-e")
                    ]
                )

    def set_battery_properties(self) -> None:
        """
        Calculate mass and power of batteries.
        :return:
        """

        self["battery cell mass"] = (
            self["energy battery mass"] * self["battery cell mass share"]
        )

        self["battery BoP mass"] = self["energy battery mass"] * (
            np.array(1.0) - self["battery cell mass share"]
        )

    def override_battery_capacity(self) -> None:
        """
        Override battery capacity.
        :return:
        """

        for key, val in self.energy_storage["capacity"].items():
            pwt, size, year = key
            if val:
                self.array.loc[
                    dict(
                        parameter="energy battery mass",
                        powertrain=pwt,
                        size=size,
                        year=year,
                    )
                ] = (
                    val
                    / self.array.loc[
                        dict(
                            parameter="battery cell energy density",
                            powertrain=pwt,
                            size=size,
                            year=year,
                        )
                    ]
                    / self.array.loc[
                        dict(
                            parameter="battery cell mass share",
                            powertrain=pwt,
                            size=size,
                            year=year,
                        )
                    ]
                )

        self.set_battery_properties()

    def override_range(self):
        """
        Set storage size or range for each powertrain.
        :return:
        """

        if self.target_range:
            for key, val in self.target_range.items():
                pwt, size, year = key

                if pwt == "BEV" and val is not None:
                    battery_DoD = self.array.loc[
                        dict(
                            powertrain=pwt,
                            size=size,
                            year=year,
                            parameter="battery DoD",
                        )
                    ]  # maximum depth of discharge allowed (80%)
                    TtW = self.array.loc[
                        dict(
                            powertrain=pwt, size=size, year=year, parameter="TtW energy"
                        )
                    ]  # kj/km

                    energy_stored = val * (TtW / battery_DoD / 3600)

                    self.array.loc[
                        dict(
                            parameter="energy battery mass",
                            powertrain=pwt,
                            size=size,
                            year=year,
                        )
                    ] = (
                        np.array(energy_stored)
                        / self.array.loc[
                            dict(
                                parameter="battery cell energy density",
                                powertrain=pwt,
                                size=size,
                                year=year,
                            )
                        ]
                        / self.array.loc[
                            dict(
                                parameter="battery cell mass share",
                                powertrain=pwt,
                                size=size,
                                year=year,
                            )
                        ]
                    )

            self.set_battery_properties()
            self.set_energy_stored_properties()
            self.set_range()

    def set_range(self) -> None:
        """
        Calculate range autonomy of vehicles
        :return:
        """

        self["range"] = (
            self["fuel mass"] * self["LHV fuel MJ per kg"] * np.array(1000)
        ) / self["TtW energy"]

        self["range"] += (
            self["electric energy stored"]
            * self["battery DoD"]
            * np.array(3600)
            / self["TtW energy"]
        )

    def check_fuel_blend(self, fuel_blend: dict) -> dict:
        for fuel, specs in fuel_blend.items():
            if "primary" not in specs:
                raise ValueError(f"Primary fuel not specified for {fuel}")

            primary = specs["primary"]

            if "share" not in primary:
                raise ValueError(f"Primary fuel share not specified for {fuel}")

            if not isinstance(primary["share"], np.ndarray):
                primary["share"] = np.array(primary["share"])

            if "type" not in primary:
                raise ValueError(f"Primary fuel type not specified for {fuel}")

            primary.setdefault(
                "name", tuple(self.bs.fuel_specs[primary["type"]]["name"])
            )
            primary.setdefault("CO2", self.bs.fuel_specs[primary["type"]]["co2"])
            primary.setdefault(
                "biogenic share", self.bs.fuel_specs[primary["type"]]["biogenic_share"]
            )

            secondary = specs.get(
                "secondary",
                {
                    "type": primary["type"],
                    "share": np.array([1]) - primary["share"],
                },
            )
            specs["secondary"] = secondary

            if "share" not in secondary:
                raise ValueError(f"Secondary fuel share not specified for {fuel}")

            if not isinstance(secondary["share"], np.ndarray):
                secondary["share"] = np.array(secondary["share"])

            if "type" not in secondary:
                raise ValueError(f"Secondary fuel type not specified for {fuel}")

            secondary.setdefault(
                "name", tuple(self.bs.fuel_specs[secondary["type"]]["name"])
            )
            secondary.setdefault("CO2", self.bs.fuel_specs[secondary["type"]]["co2"])
            secondary.setdefault(
                "biogenic share",
                self.bs.fuel_specs[secondary["type"]]["biogenic_share"],
            )

        return fuel_blend

    def set_average_lhv(self) -> None:
        """
        Calculate average LHV of fuel.
        :return:
        """

        d_map_fuel = {
            "ICEV-p": "petrol",
            "ICEV-d": "diesel",
            "HEV-d": "diesel",
            "HEV-p": "petrol",
            "PHEV-c-d": "diesel",
            "PHEV-c-p": "petrol",
            "ICEV-g": "cng",
            "FCEV": "hydrogen",
        }

        for pt in [
            pwt
            for pwt in [
                "ICEV-p",
                "ICEV-d",
                "HEV-d",
                "HEV-p",
                "PHEV-c-d",
                "PHEV-c-p",
                "ICEV-g",
                "FCEV",
            ]
            if pwt in self.array.coords["powertrain"].values
        ]:
            # calculate the average LHV based on fuel blend
            fuel_type = d_map_fuel[pt]
            primary_name = self.fuel_blend[fuel_type]["primary"]["type"]
            primary_fuel_share = self.fuel_blend[fuel_type]["primary"]["share"]
            primary_fuel_lhv = self.fuel_blend[fuel_type]["primary"].get(
                "lhv", self.bs.fuel_specs[primary_name]["lhv"]
            )
            primary_fuel_density = self.fuel_blend[fuel_type]["primary"].get(
                "density", self.bs.fuel_specs[primary_name]["density"]
            )

            if "secondary" in self.fuel_blend[fuel_type]:
                secondary_name = self.fuel_blend[fuel_type]["secondary"]["type"]
                secondary_fuel_share = self.fuel_blend[fuel_type]["secondary"]["share"]
                secondary_fuel_lhv = self.fuel_blend[fuel_type]["secondary"].get(
                    "lhv", self.bs.fuel_specs[secondary_name]["lhv"]
                )
                secondary_fuel_density = self.fuel_blend[fuel_type]["secondary"].get(
                    "density", self.bs.fuel_specs[secondary_name]["density"]
                )
            else:
                secondary_fuel_share = 0
                secondary_fuel_lhv = 0
                secondary_fuel_density = 0

            self.array.loc[dict(powertrain=pt, parameter="LHV fuel MJ per kg")] = (
                (np.array(primary_fuel_share) * primary_fuel_lhv)
                + (np.array(secondary_fuel_share) * secondary_fuel_lhv)
            ).reshape(1, -1, 1)

            self.array.loc[dict(powertrain=pt, parameter="fuel density per kg")] = (
                (np.array(primary_fuel_share) * primary_fuel_density)
                + (np.array(secondary_fuel_share) * secondary_fuel_density)
            ).reshape(1, -1, 1)

    def set_energy_stored_properties(self) -> None:
        """
        Calculate size and capacity of onboard
        energy storage components.
        :return:
        """

        self.set_average_lhv()
        self["oxidation energy stored"] = (
            self["fuel mass"] * self["LHV fuel MJ per kg"] / 3.6
        )

        self["fuel tank mass"] = (
            self["oxidation energy stored"] * self["fuel tank mass per energy"]
        )

        if "ICEV-g" in self.array.coords["powertrain"].values:
            self["fuel tank mass"] += (
                self["oxidation energy stored"] * self["CNG tank mass slope"]
                + self["CNG tank mass intercept"]
            )

        self["electric energy stored"] = (
            self["battery cell mass"] * self["battery cell energy density"]
        )

    def set_power_battery_properties(self):
        _ = lambda x: np.where(x == 0, 1, x)

        self["battery power"] = self["electric power"] * (
            self["combustion power share"] > 0
        )

        self["battery cell mass"] += (
            self["battery power"]
            / _(self["battery cell power density"])
            * (self["combustion power share"] > 0)
        )

        self["battery BoP mass"] += (
            self["battery cell mass"]
            * (np.array(1) - self["battery cell mass share"])
            * (self["combustion power share"] > 0)
        )

    def set_cargo_mass_and_annual_mileage(self):
        pass

    def set_costs(self) -> None:
        """
        Calculate the different cost types.
        :return:
        """
        self["glider cost"] = (
            self["glider base mass"] * self["glider cost slope"]
            + self["glider cost intercept"]
        )
        self["lightweighting cost"] = (
            self["glider base mass"]
            * self["lightweighting"]
            * self["glider lightweighting cost per kg"]
        )
        self["electric powertrain cost"] = (
            self["electric powertrain cost per kW"] * self["electric power"]
        )
        self["combustion powertrain cost"] = (
            self["combustion power"] * self["combustion powertrain cost per kW"]
        )
        self["fuel cell cost"] = self["fuel cell power"] * self["fuel cell cost per kW"]
        self["power battery cost"] = (
            self["battery power"] * self["power battery cost per kW"]
        )
        self["energy battery cost"] = (
            self["energy battery cost per kWh"] * self["electric energy stored"]
        )
        self["fuel tank cost"] = self["fuel tank cost per kg"] * self["fuel mass"]
        # Per km
        self["energy cost"] = self["energy cost per kWh"] * self["TtW energy"] / 3600

        # For battery, need to divide cost of electricity
        # at battery by efficiency of charging
        # to get costs at the "wall socket".

        _ = lambda x: np.where(x == 0, 1, x)
        self["energy cost"] /= _(self["battery charge efficiency"])

        self["component replacement cost"] = (
            self["energy battery cost"] * self["battery lifetime replacements"]
            + self["fuel cell cost"] * self["fuel cell lifetime replacements"]
        )

        with open(self.DATA_DIR / "purchase_cost_params.yaml", "r") as stream:
            to_markup = yaml.safe_load(stream)["markup"]

        self[to_markup] *= self["markup factor"]

        # calculate costs per km:
        self["lifetime"] = self["lifetime kilometers"] / self["kilometers per year"]

        with open(self.DATA_DIR / "purchase_cost_params.yaml", "r") as stream:
            purchase_cost_params = yaml.safe_load(stream)["purchase"]

        self["purchase cost"] = self[purchase_cost_params].sum(axis=2)
        # per km
        amortisation_factor = self["interest rate"] + (
            self["interest rate"]
            / (
                (np.array(1) + self["interest rate"]) ** self["lifetime kilometers"]
                - np.array(1)
            )
        )
        self["amortised purchase cost"] = (
            self["purchase cost"] * amortisation_factor / self["kilometers per year"]
        )

        # per km
        self["maintenance cost"] = (
            self["maintenance cost per glider cost"]
            * self["glider cost"]
            / self["kilometers per year"]
        )

        # simple assumption that component replacement
        # occurs at half of life.
        self["amortised component replacement cost"] = (
            (
                self["component replacement cost"]
                * (
                    (np.array(1) - self["interest rate"]) ** self["lifetime kilometers"]
                    / 2
                )
            )
            * amortisation_factor
            / self["kilometers per year"]
        )

        self["total cost per km"] = (
            self["energy cost"]
            + self["amortised purchase cost"]
            + self["maintenance cost"]
            + self["amortised component replacement cost"]
        )

    def set_ttw_efficiency(self) -> None:
        """
        Fill in the tank-to-wheel efficiency
        calculated by `calculate_ttw_efficiency`.
        :return:
        """
        _ = lambda array: np.where(array == 0, 1, array)

        self["TtW efficiency"] = (
            _(self["fuel cell system efficiency"])
            * self["transmission efficiency"]
            * self["engine efficiency"]
        )

        self["TtW efficiency"] *= np.where(
            self["charger mass"] > 0, self["battery discharge efficiency"], 1
        )

    def set_hot_emissions(self) -> None:
        """
        Calculate hot pollutant emissions based on ``driving cycle``.
        The driving cycle is passed to the :class:`HotEmissionsModel` class
        and :meth:`get_emissions_per_powertrain`
        return emissions per substance per second of driving cycle.
        Those are summed up and divided by
        the distance driven, to obtain emissions, in kg per km.
        :return: Does not return anything. Modifies ``self.array`` in place.
        """

        hem = HotEmissionsModel(
            velocity=self.energy.sel(parameter="velocity"),
            cycle_name=self.cycle,
            vehicle_type=self.vehicle_type,
            powertrains=self.array.coords["powertrain"].values,
            sizes=self.array.coords["size"].values,
        )

        with open(
            self.DATA_DIR / "emission_factors" / "exhaust_flows.yaml", "r"
        ) as stream:
            list_direct_emissions = sorted(yaml.safe_load(stream))

        list_direct_emissions = [
            e + f", {c}"
            for c in ["urban", "suburban", "rural"]
            for e in list_direct_emissions
        ]

        with open(
            self.DATA_DIR / "emission_factors" / "euro_classes.yaml", "r"
        ) as stream:
            euro_classes = yaml.safe_load(stream)[self.vehicle_type]

        list_years = np.clip(
            self.array.coords["year"].values,
            min(euro_classes.keys()),
            max(euro_classes.keys()),
        )

        list_euro_classes = [euro_classes[y] for y in list(list_years)]

        # to calculate emissions and degradation factors
        # we need the vehicle's lifetime, annual mileage
        # as well as its instant fuel consumption

        energy_consumption = self.energy.sel(
            parameter=["motive energy", "auxiliary energy", "recuperated energy"],
            size=self.array.coords["size"].values,
            powertrain=self.array.coords["powertrain"].values,
        ).sum(dim="parameter")

        hot_emissions = hem.get_hot_emissions(
            euro_class=list_euro_classes,
            lifetime_km=self["lifetime kilometers"],
            energy_consumption=energy_consumption,
            yearly_km=self["kilometers per year"],
        ).values

        self.array.loc[
            dict(
                parameter=list_direct_emissions,
            )
        ] = hot_emissions

    def set_particulates_emission(self) -> None:
        """
        Calculate the emission of particulates according to
        https://www.eea.europa.eu/ds_resolveuid/6USNA27I4D

        and further disaggregated in:
        https://doi.org/10.1016/j.atmosenv.2020.117886

        for:

        - brake wear
        - tire wear
        - road wear
        - re-suspended road dust

        by considering:

        - vehicle mass
        - driving situation (urban, rural, motorway)

        into the following fractions:

        - PM 2.5
        - PM 10

        Emissions are subdivided in compartments: urban, suburban and rural.

        """

        list_param = [
            "tire wear emissions",
            "brake wear emissions",
            "road wear emissions",
            "road dust emissions",
        ]

        pem = ParticulatesEmissionsModel(
            velocity=self.energy.sel(parameter="velocity"),
            mass=self["driving mass"],
        )

        self[list_param] = pem.get_abrasion_emissions()

        # brake emissions are discounted by
        # the use of regenerative braking
        self["brake wear emissions"] *= np.array(1) - self["share recuperated energy"]

    def set_noise_emissions(self) -> None:
        """
        Calculate noise emissions based on ``driving cycle``.
        The driving cycle is passed to the :class:`NoiseEmissionsModel` class
        and :meth:`get_sound_power_per_compartment`
        returns emissions per compartment type ("rural", "non-urban" and "urban")
        per second of driving cycle.

        Noise emissions are not differentiated by size classes at the moment,
        but only by powertrain "type"
        (e.g., combustion, hybrid and electric)

        :return: Does not return anything. Modifies ``self.array`` in place.
        """
        velocity = self.energy.sel(parameter="velocity")
        nem = NoiseEmissionsModel(velocity, vehicle_type=self.vehicle_type)

        with open(
            self.DATA_DIR / "emission_factors" / "noise_flows.yaml", "r"
        ) as stream:
            list_noise_emissions = yaml.safe_load(stream)

        self.array.loc[
            dict(parameter=list_noise_emissions)
        ] = nem.get_sound_power_per_compartment()

    def calculate_cost_impacts(self, sensitivity=False) -> xr.DataArray:
        """
        This method returns an array with cost values per vehicle-km,
        subdivided into the following groups:

            * Purchase
            * Maintenance
            * Component replacement
            * Energy
            * Total cost of ownership

        :return: A xarray array with cost information per vehicle-km
        """

        list_cost_cat = [
            "purchase",
            "maintenance",
            "component replacement",
            "energy",
            "total",
        ]

        response = self.array.sel(
            parameter=[
                "amortised purchase cost",
                "maintenance cost",
                "amortised component replacement cost",
                "energy cost",
                "total cost per km",
            ],
        )

        response.coords["parameter"] = list_cost_cat

        if not sensitivity:
            return response

        return response / response.sel(value="reference")



"""
hot_emissions.py contains HotEmissionModel
which calculates fuel-related exhaust emissions.
"""

from pathlib import Path
from typing import Any, List, Union

import numpy as np
import pandas as pd
import xarray as xr
import yaml
from xarray import DataArray

from . import DATA_DIR

FILEPATH_DC_SPECS = DATA_DIR / "driving cycle" / "dc_specs.yaml"

MAP_PWT = {
    "Human": "BEV",
    "BEV": "BEV",
    "BEV-depot": "BEV",
    "BEV-opp": "BEV",
    "BEV-motion": "BEV",
    "PHEV-e": "BEV",
    "PHEV-c-d": "ICEV-d",
    "PHEV-d": "ICEV-d",
    "PHEV-c-p": "ICEV-p",
    "PHEV-p": "ICEV-p",
    "ICEV-d": "ICEV-d",
    "ICEV-g": "ICEV-g",
    "ICEV-p": "ICEV-p",
    "HEV-p": "ICEV-p",
    "HEV-d": "ICEV-d",
    "FCEV": "BEV",
}


def _(obj: Union[np.ndarray, xr.DataArray]) -> Union[np.ndarray, xr.DataArray]:
    """Add a trailing dimension to make input arrays broadcast correctly"""
    if isinstance(obj, (np.ndarray, xr.DataArray)):
        return np.expand_dims(obj, -1)
    return obj


def get_emission_factors(filepath) -> [Any, None]:
    """Hot emissions factors extracted for passenger cars from HBEFA 4.1
    detailed by size, powertrain and EURO class for each substance.
    """

    try:
        df = pd.read_csv(filepath, sep=",")
        cols = ["powertrain", "component"]

        if "euro_class" in df.columns:
            cols.append("euro_class")

        if "size" in df.columns:
            cols.insert(1, "size")
        if "type" in df.columns:
            cols.insert(-2, "type")

        ef = df.groupby(cols)["value"].mean().to_xarray()

    except FileNotFoundError:
        return None

    return ef.fillna(0.0)


def get_mileage_degradation_factor(
    lifetime_km: xr.DataArray,
    euro_class: List[int],
    powertrains: List[str],
    vehicle_type: str,
) -> [DataArray, None]:
    """
    Catalyst degrade overtime, leading to increased emissions
    of CO, HC and NOX. We apply a correction factor from HBEFA 4.1
    to reflect this.
    :return:
    """

    corr = get_emission_factors(
        filepath=DATA_DIR / "emission_factors" / vehicle_type / f"degradation_EF.csv",
    )

    if corr is None:
        return None

    corr = corr.fillna(1.0)
    corr = corr.expand_dims({"km": 2}).copy()

    if vehicle_type == "car":
        max_km = 200000
    elif vehicle_type in ["bus", "truck"]:
        max_km = 890000
    else:
        max_km = 200000

    corr = corr.assign_coords({"km": np.array([0, max_km])})

    corr = corr.sel(
        powertrain=[
            p for p in lifetime_km.powertrain.values if p in corr.powertrain.values
        ],
        euro_class=euro_class,
    )

    corr = corr.expand_dims({"size": len(lifetime_km.coords["size"])})
    corr = corr.assign_coords({"size": lifetime_km.coords["size"].values})

    corr = corr.transpose("euro_class", "powertrain", "size", "component", "km")

    corr = corr.interp(
        km=lifetime_km.max(), method="linear", kwargs={"fill_value": "extrapolate"}
    )

    corr.values = np.where(corr < 1, 1.0, corr)

    return corr


def get_driving_cycle_compartments(cycle_name, vehicle_type) -> dict:
    with open(FILEPATH_DC_SPECS, "r") as f:
        return yaml.safe_load(f)["environments"][vehicle_type][cycle_name]


class HotEmissionsModel:
    """
    Calculate hot pollutants emissions based on HBEFA 4.1 data, function of fuel consumption
    for vehicles with a combustion engine.

    :param cycle: Driving cycle. Pandas Series of second-by-second speeds (km/h) or name (str)
        of cycle e.g., "WLTC","WLTC 3.1","WLTC 3.2","WLTC 3.3","WLTC 3.4","CADC Urban","CADC Road",
        "CADC Motorway","CADC Motorway 130","CADC","NEDC".
    :type cycle: pandas.Series

    """

    def __init__(
        self,
        powertrains: np.array,
        sizes: np.array,
        velocity: np.ndarray,
        cycle_name: str,
        vehicle_type: str,
    ) -> None:
        self.powertrains = powertrains
        self.sizes = sizes
        self.velocity = velocity / 1000 * 3600  # m/s to km/h
        self.cycle_name = cycle_name
        self.vehicle_type = vehicle_type
        self.exhaust = get_emission_factors(
            filepath=DATA_DIR
            / "emission_factors"
            / vehicle_type
            / f"EF_HBEFA42_exhaust.csv"
        )
        self.non_exhaust = get_emission_factors(
            filepath=DATA_DIR
            / "emission_factors"
            / vehicle_type
            / f"EF_HBEFA42_non_exhaust.csv"
        )
        self.nmhc_species = get_emission_factors(
            filepath=DATA_DIR / "emission_factors" / vehicle_type / f"NMHC_species.csv"
        )
        self.engine_wear = get_emission_factors(
            filepath=DATA_DIR / "emission_factors" / vehicle_type / f"engine_wear.csv"
        )

    def get_hot_emissions(
        self,
        euro_class: List[int],
        lifetime_km: xr.DataArray,
        energy_consumption: xr.DataArray,
        yearly_km: xr.DataArray,
    ) -> DataArray:
        """
        Calculate hot pollutants emissions given a powertrain type (i.e., diesel, petrol, CNG)
        and a EURO pollution class, per air sub-compartment (i.e., urban, suburban and rural).
        Note that Nh3 and N2O emissions do not depend on the speed level. For those, average
        values observed across different traffic situations are used instead.
        Also includes cold start emissions. Cold start emissions are also given by HBEFA 4.1
        and expressed in given in g/start. Cold start emissions are divided by the average trip
        length in Europe (20 km), to normalize them per km.

        The emission sums are further divided into `air compartments`: urban, suburban and rural.

            * *urban*: from 0 to 50 km/k
            * *suburban*: from 51 km/h to 80 km/h
            * *rural*: above 80 km/h

        :param powertrain_type: "diesel", "petrol" or "CNG"
        :param euro_class: integer, corresponding to the EURO pollution class
        :param energy_consumption: tank-to-wheel energy consumption for each second of the driving cycle
        :param yearly_km: annual mileage, to calculate cold start emissions
        :return: Pollutants emission per km driven, per air compartment.
        """

        hot_emissions = self.exhaust.sel(
            powertrain=[
                MAP_PWT[pt] if MAP_PWT[pt] in self.exhaust.powertrain.values else "BEV"
                for pt in lifetime_km.coords["powertrain"].values
            ],
            euro_class=euro_class,
        )

        if "size" not in hot_emissions.dims:
            hot_emissions = hot_emissions.expand_dims(
                {"size": len(lifetime_km.coords["size"])}
            )
            hot_emissions.coords["size"] = lifetime_km.coords["size"]

        hot_emissions = hot_emissions.transpose(
            "euro_class", "powertrain", "size", "component"
        )

        hot_emissions = hot_emissions.sel(size=lifetime_km.coords["size"].values)

        energy_consumption = energy_consumption.sel(
            size=lifetime_km.coords["size"].values,
            powertrain=lifetime_km.coords["powertrain"].values,
            year=lifetime_km.coords["year"].values,
        )

        distance = self.velocity.sum(dim="second") / 3600

        # Emissions for each second of the driving cycle equal:
        # a * energy consumption
        # with a being a coefficient given by fitting HBEFA 4.1 data
        # the fitting of emissions function of energy consumption
        # is described in the notebook
        # `HBEFA trucks.ipynb` in the folder `dev`.

        # energy consumption is given in kj for each second
        # emissions are in grams per MJ

        emissions = xr.DataArray(
            hot_emissions.values * _(energy_consumption / 1000),
            dims=["second", "value", "year", "powertrain", "size", "component"],
            coords={
                "second": energy_consumption.coords["second"],
                "value": energy_consumption.coords["value"],
                "year": energy_consumption.coords["year"],
                "powertrain": energy_consumption.coords["powertrain"],
                "size": energy_consumption.coords["size"],
                "component": hot_emissions.coords["component"],
            },
        )

        # a bit of a manual calibration for N2O and NH3
        # as they do not correlate with fuel consumption

        if self.vehicle_type == "car":
            emissions.loc[dict(component="Dinitrogen oxide")] *= 0.5
            emissions.loc[dict(component="Ammonia")] *= 0.5
        elif self.vehicle_type == "truck":
            emissions.loc[dict(component="Dinitrogen oxide")] *= 10
            emissions.loc[dict(component="Ammonia")] *= 10
        elif self.vehicle_type == "bus":
            emissions.loc[dict(component="Dinitrogen oxide")] /= 15
            emissions.loc[dict(component="Ammonia")] /= 12
        else:
            pass

        # apply a mileage degradation factor for CO, HC and NOx
        degradation_correction = get_mileage_degradation_factor(
            lifetime_km=lifetime_km,
            euro_class=euro_class,
            powertrains=emissions.powertrain.values,
            vehicle_type=self.vehicle_type,
        )

        if degradation_correction is not None:
            emissions.loc[
                dict(
                    powertrain=degradation_correction.powertrain.values,
                    component=degradation_correction.component.values,
                )
            ] *= degradation_correction.values

        if self.non_exhaust is not None:
            non_exhaust = self.non_exhaust.sel(
                powertrain=[
                    MAP_PWT[pt]
                    if MAP_PWT[pt] in self.non_exhaust.powertrain.values
                    else "BEV"
                    for pt in emissions.powertrain.values
                ],
                euro_class=euro_class,
                component=[
                    c
                    for c in emissions.component.values
                    if c in self.non_exhaust.component.values
                ],
            )

            if "size" not in non_exhaust.dims:
                non_exhaust = non_exhaust.expand_dims(
                    {"size": len(energy_consumption.coords["size"])}
                )
                non_exhaust.coords["size"] = energy_consumption.coords["size"]

            non_exhaust = non_exhaust.transpose(
                "euro_class", "powertrain", "size", "component", "type"
            )

            non_exhaust = non_exhaust.sel(size=lifetime_km.coords["size"].values)

            start_per_day = 2.3  # source for

            # Cold start and soak emissions are defined per start and stop
            # Therefore, we need to normalize per km
            # And add cold start emissions to the first second of the driving cycle

            yearly_km = yearly_km.transpose("value", "year", "powertrain", "size")

            emissions.loc[
                dict(
                    second=emissions.second.values[0],
                    component=non_exhaust.component.values,
                )
            ] += (
                _(distance / yearly_km * start_per_day * 365)
                * non_exhaust.sel(type="cold start").values
            )

            # And add soak emissions to the last second of the driving cycle
            emissions.loc[
                dict(
                    second=emissions.second.values[-1],
                    component=non_exhaust.component.values,
                )
            ] += (
                _(distance / yearly_km * start_per_day * 365)
                * non_exhaust.loc[dict(type="soak")].values
            )

            # Diurnal emissions are defined in g/day
            # And need to be evenly distributed
            # throughout the driving cycle

            daily_km_to_year = distance / (yearly_km / 365)

            emissions.loc[dict(component=non_exhaust.component.values)] += (
                _(daily_km_to_year)
                * non_exhaust.sel(type="diurnal").values
                / emissions.shape[0]
            )

            # Running losses are in g/km (no conversion needed)
            # And need to be evenly distributed throughout the driving cycle

            emissions.loc[dict(component=non_exhaust.component.values)] += (
                _(distance)
                * non_exhaust.sel(type="running losses").values
                / emissions.shape[0]
            )

        # Add additional species derived from NMHC emissions
        # Toluene, Xylene, Formaldehyde, Acetaldehyde, etc.
        # Also heavy metals
        # And add them to the emissions array

        nmhc = self.nmhc_species.sel(
            powertrain=[
                MAP_PWT[pt] if pt in self.nmhc_species.powertrain.values else "BEV"
                for pt in emissions.powertrain.values
            ]
        )
        nmhc = nmhc.assign_coords({"powertrain": emissions.powertrain.values})

        if "size" not in nmhc.dims:
            nmhc = nmhc.expand_dims({"size": len(energy_consumption.coords["size"])})
            nmhc.coords["size"] = energy_consumption.coords["size"]

        if "year" not in nmhc.dims:
            nmhc = nmhc.expand_dims({"year": len(energy_consumption.coords["year"])})
            nmhc.coords["year"] = energy_consumption.coords["year"]

        nmhc = nmhc.transpose(
            "year",
            "powertrain",
            "size",
            "component",
        )
        nmhc = nmhc.sel(size=energy_consumption.coords["size"].values)

        emissions.loc[dict(component=nmhc.coords["component"].values)] = (
            nmhc.values * emissions.sel(component=["Non-methane hydrocarbon"]).values
        )

        emissions.loc[dict(component="Non-methane hydrocarbon")] *= nmhc.sum(
            dim="component"
        ).values

        # Heavy metals emissions are dependent of fuel consumption
        # given in grams of emission per kj

        engine_wear = self.engine_wear.sel(
            powertrain=[
                MAP_PWT[pt]
                if MAP_PWT[pt] in self.engine_wear.powertrain.values
                else "BEV"
                for pt in emissions.powertrain.values
            ]
        )
        engine_wear = engine_wear.assign_coords(
            {"powertrain": emissions.powertrain.values}
        )

        if "size" not in engine_wear.dims:
            engine_wear = engine_wear.expand_dims(
                {"size": len(energy_consumption.coords["size"])}
            )
            engine_wear.coords["size"] = energy_consumption.coords["size"]

        if "year" not in engine_wear.dims:
            engine_wear = engine_wear.expand_dims(
                {"year": len(energy_consumption.coords["year"])}
            )
            engine_wear.coords["year"] = energy_consumption.coords["year"]

        engine_wear = engine_wear.transpose(
            "year",
            "powertrain",
            "size",
            "component",
        )

        emissions.loc[dict(component=engine_wear.coords["component"].values)] += (
            energy_consumption * engine_wear
        ).values

        # reorder the component coordinates by alphabetical order
        emissions = emissions.reindex(component=sorted(emissions.component.values))

        # urban emissions are the sum of emissions
        # along the ``second`` dimension
        # where velocity is below 50 km/h
        # converted to kg/km
        emissions = emissions.fillna(0)

        urban_emissions = (
            np.where(_(self.velocity) <= 50, emissions, 0).sum(axis=0)
            / _(distance)
            / 1000
        )

        # rural emissions are the sum of emissions
        # along the ``second`` dimension
        # where velocity is between 50 km/h
        # and 80 km/h
        rural_emissions = (
            np.where(
                (_(self.velocity) > 50) & (_(self.velocity) <= 80), emissions, 0
            ).sum(axis=0)
            / _(distance)
            / 1000
        )

        # highway emissions are the sum of emissions
        # along the ``second`` dimension
        # where velocity is above 80 km/h
        highway_emissions = (
            np.where(_(self.velocity) > 80, emissions, 0).sum(axis=0)
            / _(distance)
            / 1000
        )

        list_direct_emissions = [
            e + f", {c}"
            for c in ["urban", "suburban", "rural"]
            for e in emissions.coords["component"].values
        ]

        res = xr.DataArray(
            np.concatenate(
                (urban_emissions, rural_emissions, highway_emissions), axis=-1
            ),
            coords=[
                emissions.coords["value"],
                emissions.coords["year"],
                emissions.coords["powertrain"],
                emissions.coords["size"],
                list_direct_emissions,
            ],
            dims=["value", "year", "powertrain", "size", "component"],
        )

        return res.transpose("size", "powertrain", "component", "year", "value")



import itertools

import numpy as np
import pandas as pd
import stats_arrays as sa
import xarray as xr

from .vehicle_input_parameters import VehicleInputParameters as vip


def fill_xarray_from_input_parameters(input_parameters, sensitivity=False, scope=None):
    """Create an `xarray` labeled array from the sampled input parameters.


    This function extracts the parameters' names and values contained in the
    `parameters` attribute of the :class:`CarInputParameters` class
    in :mod:`car_input_parameters` and insert them into a
    multi-dimensional numpy-like array from the *xarray* package
    (http://xarray.pydata.org/en/stable/).


    :param sensitivity:
    :param input_parameters: Instance of the :class:`TruckInputParameters` class
    in :mod:`truck_input_parameters`.
    :returns: `tuple`, `xarray.DataArray`
    - tuple (`size_dict`, `powertrain_dict`, `parameter_dict`, `year_dict`)
    - array

    Dimensions of `array`:

        0. Vehicle size, e.g. "3.5t", "7.5t", etc. str.
        1. Powertrain, e.g. "ICE-d", "BEV". str.
        2. Year. int.
        3. Samples.

    """

    # Check whether the argument passed is an instance of :class:`TruckInputParameters`
    if not isinstance(input_parameters, vip):
        raise TypeError(
            "The argument passed is not an object of the TruckInputParameter class"
        )

    if scope is None:
        scope = {
            "size": input_parameters.sizes,
            "powertrain": input_parameters.powertrains,
            "year": input_parameters.years,
        }
    else:
        if "size" not in scope:
            scope["size"] = input_parameters.sizes
        if "powertrain" not in scope:
            scope["powertrain"] = input_parameters.powertrains
        if "year" not in scope:
            scope["year"] = input_parameters.years

    # Make sure to include PHEV-e and PHEV-c-d if
    # PHEV-d is listed

    missing_pwts = [
        ("PHEV-d", "PHEV-e", "PHEV-c-d"),
        ("PHEV-p", "PHEV-e", "PHEV-c-p"),
    ]

    for missing_pwt in missing_pwts:
        if missing_pwt[0] in scope["powertrain"]:
            if not any(p in scope["powertrain"] for p in missing_pwt[1:]):
                scope["powertrain"].extend(missing_pwt[1:])

    if any(s for s in scope["size"] if s not in input_parameters.sizes):
        raise ValueError("One of the size types is not valid.")

    if any(y for y in scope["year"] if y not in input_parameters.years):
        raise ValueError("One of the years defined is not valid.")

    if any(pt for pt in scope["powertrain"] if pt not in input_parameters.powertrains):
        raise ValueError("One of the powertrain types is not valid.")

    # if the purpose is not to do a sensitivity analysis
    # the dimension `value` of the array is as large as
    # the number of iterations to perform
    # that is, 1 in `static` mode, or several in `stochastic` mode.

    d = {v: k for k, v in enumerate(scope["size"])}

    if not sensitivity:
        array = xr.DataArray(
            np.zeros(
                (
                    len(scope["size"]),
                    len(scope["powertrain"]),
                    len(input_parameters.parameters),
                    len(scope["year"]),
                    input_parameters.iterations or 1,
                )
            ),
            coords=[
                sorted(scope["size"], key=lambda x: d[x]),
                scope["powertrain"],
                input_parameters.parameters,
                scope["year"],
                np.arange(input_parameters.iterations or 1),
            ],
            dims=["size", "powertrain", "parameter", "year", "value"],
        ).astype("float32")

    # if the purpose is to do a sensitivity analysis
    # then the length of the dimensions `value` equals the number of parameters
    else:
        params = ["reference"]
        params.extend([a for a in input_parameters.input_parameters])
        array = xr.DataArray(
            np.zeros(
                (
                    len(scope["size"]),
                    len(scope["powertrain"]),
                    len(input_parameters.parameters),
                    len(scope["year"]),
                    len(params),
                )
            ),
            coords=[
                scope["size"],
                scope["powertrain"],
                input_parameters.parameters,
                scope["year"],
                params,
            ],
            dims=["size", "powertrain", "parameter", "year", "value"],
        ).astype("float32")

    size_dict = {k: i for i, k in enumerate(scope["size"])}
    powertrain_dict = {k: i for i, k in enumerate(scope["powertrain"])}
    year_dict = {k: i for i, k in enumerate(scope["year"])}
    parameter_dict = {k: i for i, k in enumerate(input_parameters.parameters)}

    for param in input_parameters:
        pwt = (
            set(input_parameters.metadata[param]["powertrain"])
            if isinstance(input_parameters.metadata[param]["powertrain"], list)
            else set([input_parameters.metadata[param]["powertrain"]])
        )

        size = (
            set(input_parameters.metadata[param]["sizes"])
            if isinstance(input_parameters.metadata[param]["sizes"], list)
            else set([input_parameters.metadata[param]["sizes"]])
        )

        year = (
            set(input_parameters.metadata[param]["year"])
            if isinstance(input_parameters.metadata[param]["year"], list)
            else set([input_parameters.metadata[param]["year"]])
        )

        if (
            pwt.intersection(scope["powertrain"])
            and size.intersection(scope["size"])
            and year.intersection(scope["year"])
        ):
            array.loc[
                dict(
                    powertrain=[p for p in pwt if p in scope["powertrain"]],
                    size=[s for s in size if s in scope["size"]],
                    year=[y for y in year if y in scope["year"]],
                    parameter=input_parameters.metadata[param]["name"],
                )
            ] = input_parameters.values[param]

    if sensitivity:
        # we increase each value by 10%

        list_params = list(set([a for a in input_parameters.input_parameters]))

        for param in list_params:
            array.loc[dict(parameter=param, value=param)] *= 1.1

    return (size_dict, powertrain_dict, parameter_dict, year_dict), array
